exact_fetched = {
    "aerospike": "https://github.com/aerospike/aerospike-client-python",
    "appier": "https://github.com/hivesolutions/appier",
    "appium-python-client": "https://github.com/appium/python-client",
    "boxsdk": "https://github.com/box/box-python-sdk",
    "breezy": "https://github.com/breezy-team/breezy",  # should be https://launchpad.net/brz, but that has a higher distance
    "django-lfs": "https://github.com/diefenbach/django-lfs",
    "giacpy": "https://gitlab.math.univ-paris-diderot.fr/han/giacpy",
    "github-api-python": "https://github.com/hivesolutions/github_api",
    "libvirt-python": "https://gitlab.com/libvirt/libvirt-python",
    "mygpoclient": "https://github.com/gpodder/mygpoclient",
    "oslo-upgradecheck": "https://opendev.org/openstack/oslo.upgradecheck",
    "oslo-vmware": "https://opendev.org/openstack/oslo.vmware",
    "pycli": "https://github.com/whilp/cli",
    "pycryptodomex": "https://github.com/Legrandin/pycryptodome",
    "pynliner": "https://github.com/rennat/pynliner",
    "pyriemann": "https://github.com/alexandrebarachant/pyRiemann",
    "python-nss": "https://hg.mozilla.org/projects/python-nss",
    "python-termstyle": "https://github.com/gfxmonk/termstyle",
    "pywurfl": "https://sourceforge.net/projects/wurfl",
    "PyZ3950": "https://www.panix.com/~asl2/software/PyZ3950",
    "roundup": "https://sourceforge.net/projects/roundup",
    "rpy2": "https://github.com/rpy2/rpy2",
    "South": "https://bitbucket.org/andrewgodwin/south",
    "tinysegmenter": "http://git.tuxfamily.org/tinysegmente/tinysegmenter/",
    "webapp2": "https://github.com/GoogleCloudPlatform/webapp2",  # upgrade to https
    "web.py": "https://github.com/webpy/webpy",
}

exact_metadata = {
    "agate-excel": "https://github.com/wireservice/agate-excel",
    "alex-ber-utils": "https://github.com/alex-ber/AlexBerUtils",
    "algoliasearch": "https://github.com/algolia/algoliasearch-client-python",
    "ansible-tower-cli": "https://github.com/ansible/tower-cli",
    "apache-airflow": "https://github.com/apache/airflow",
    "apache-beam": "https://github.com/apache/beam",
    "aspy-refactor-imports": "https://github.com/asottile/aspy.refactor_imports",
    "auger-python": "https://github.com/laffra/auger",
    "backports-functools-lru-cache": "https://github.com/jaraco/backports.functools_lru_cache",
    "backports-shutil-get-terminal-size": "https://github.com/chrippa/backports.shutil_get_terminal_size",
    "backports-shutil-which": "https://github.com/minrk/backports.shutil_which",
    "backports-ssl-match-hostname": "https://bitbucket.org/brandon/backports.ssl_match_hostname",
    "betamax-serializers": "https://gitlab.com/betamax/serializers",
    "bolt-python": "https://github.com/bolt-project/bolt",
    "BytecodeAssembler": "https://github.com/peak-legacy/bytecodeassembler",
    "certbot": "https://github.com/certbot/certbot",  # TODO rename https://github.com/letsencrypt/letsencrypt
    "certifi": "https://github.com/certifi/python-certifi",
    "cffi": "https://foss.heptapod.net/pypy/cffi",  # https://foss.heptapod.net/pypy/cffi/issues/443
    "config": "https://bitbucket.org/vinay.sajip/config",
    "Contextual": "https://github.com/peak-legacy/contextual",
    "couchbase": "https://github.com/couchbase/couchbase-python-client",
    "cssutils": "https://bitbucket.org/cthedot/cssutils",
    "dbus-signature-pyparsing": "https://github.com/stratis-storage/dbus-signature-pyparsing",
    "ddtrace": "https://github.com/DataDog/dd-trace-py",
    "DecoratorTools": "https://github.com/peak-legacy/decoratortools",
    "delegator-py": "https://github.com/kennethreitz/delegator",
    "dephell-pythons": "https://github.com/dephell/dephell_pythons",
    "dephell-shells": "https://github.com/dephell/dephell_shells",
    "dephell-specifier": "https://github.com/dephell/dephell_specifier",
    "django-gravatar": "https://code.google.com/p/django-gravatar",
    "django-hijack-admin": "https://github.com/arteria/django-hijack-admin",
    "django-pagination": "https://code.google.com/p/django-pagination",
    "djangorestframework-bulk": "https://github.com/miki725/django-rest-framework-bulk",
    "djangorestframework-csv": "https://github.com/mjumbewu/django-rest-framework-csv",
    "djangorestframework-filters": "https://github.com/philipn/django-rest-framework-filters",
    "djangorestframework-gis": "https://github.com/djangonauts/django-rest-framework-gis",
    "djangorestframework-jsonapi": "https://github.com/django-json-api/django-rest-framework-json-api",
    "djangorestframework-jsonp": "https://github.com/jpadilla/django-rest-framework-jsonp",
    "djangorestframework-jwt": "https://github.com/GetBlimp/django-rest-framework-jwt",
    "djangorestframework-oauth": "https://github.com/jpadilla/django-rest-framework-oauth",
    "djangorestframework-recursive": "https://github.com/heywbj/django-rest-framework-recursive",
    "djangorestframework-sso": "https://github.com/namespace-ee/django-rest-framework-sso",
    "djangorestframework-xml": "https://github.com/jpadilla/django-rest-framework-xml",
    "djangorestframework-yaml": "https://github.com/jpadilla/django-rest-framework-yaml",
    "django-websocket": "https://github.com/gregmuellegger/django-websocket",
    "docker-pycreds": "https://github.com/shin-/dockerpy-creds",
    "dotnetcore2": "https://github.com/dotnet/core",
    "drfdocs": "https://github.com/manosim/django-rest-framework-docs",
    "drf-jwt": "https://github.com/Styria-Digital/django-rest-framework-jwt",
    "edx-ccx-keys": "https://github.com/edx/ccx-keys",
    "edx-completion": "https://github.com/edx/completion",
    "edx-i18n-tools": "https://github.com/edx/i18n-tools",
    "edx-opaque-keys": "https://github.com/edx/opaque-keys",
    "elasticsearch2": "https://github.com/elastic/elasticsearch-py",
    "empy": "http://www.alcyone.com/software/empy",
    "etcd": "https://github.com/dsoprea/PythonEtcdClient",
    "evergreen-py": "https://github.com/evergreen-ci/evergreen.py",
    "facebookads": "https://github.com/facebook/facebook-python-ads-sdk",
    "facebook-business": "https://github.com/facebook/facebook-python-business-sdk",
    "gdata": "https://github.com/google/gdata-python-client",
    "getchanges": "https://github.com/TheKevJames/getchanges",
    "girder-client": "https://github.com/girder/girder",
    "google-apitools": "https://github.com/google/apitools",
    "google-nucleus": "https://github.com/google/nucleus",
    "google-pasta": "https://github.com/google/pasta",
    "graphql-core": "https://github.com/graphql-python/graphql-core",
    "graphql-core-next": "https://github.com/graphql-python/graphql-core-next",
    "guess_language-spirit": "https://bitbucket.org/spirit/guess_language",
    "hacs-frontend": "https://github.com/hacs/frontend",
    "hdfs": "https://github.com/mtth/hdfs",
    "honeycomb-beeline": "https://github.com/honeycombio/beeline-python",
    "hs-dbus-signature": "https://github.com/stratis-storage/hs-dbus-signature",
    "ibm-cos-sdk-core": "https://github.com/ibm/ibm-cos-sdk-python-core",
    "ibm-cos-sdk-s3transfer": "https://github.com/ibm/ibm-cos-sdk-python-s3transfer",
    "ibm-db": "https://github.com/ibmdb/python-ibmdb",
    "imagecodecs": "https://github.com/cgohlke/imagecodecs",
    "Importing": "http://svn.eby-sarna.com/Importing/",
    "imreg": "https://github.com/cgohlke/imreg",
    "infi-clickhouse-orm": "https://github.com/Infinidat/infi.clickhouse_orm",
    "j1m.sphinxautointerface": "https://bitbucket.org/j1m/sphinxautointerface",
    "jobspy": "https://github.com/josiahcarlson/jobs",
    "k5test": "https://github.com/pythongssapi/k5test",
    "lancet-ioam": "https://github.com/ioam/lancet",
    "linaro-json": "https://launchpad.net/linaro-python-json",
    "localstack-client": "https://github.com/localstack/localstack-python-client",
    "logentries-lecli": "https://github.com/logentries/lecli",
    "logging": "https://bitbucket.org/vinay.sajip/logging",
    "lzmaffi": "https://github.com/r3m0t/backports.lzma",
    "magic": "http://www.jsnp.net/code/magic.py",
    "mailman-hyperkitty": "https://gitlab.com/mailman/mailman-hyperkitty/",
    "mesh-tensorflow": "https://github.com/tensorflow/mesh",
    "mockredispy": "https://github.com/locationlabs/mockredis",
    "neo4j-driver": "https://github.com/neo4j/neo4j-python-driver",
    "nexpose": "https://github.com/rapid7/nexpose-client-python",
    "nteract-scrapbook": "https://github.com/nteract/scrapbook",
    "open3d-python": "https://github.com/intel-isl/Open3D",
    "opentracing-instrumentation": "https://github.com/uber-common/opentracing-python-instrumentation",
    "pantsbuild-pants": "https://github.com/pantsbuild/pants",
    "path-py": "https://github.com/jaraco/path",
    "PEAK": "https://github.com/peak-legacy/peak",
    "peak-rules": "https://github.com/peak-legacy/peak-rules",
    "pep8": "https://github.com/PyCQA/pep8",
    "pgc-interface": "https://github.com/UniversalDevicesInc/pgc-python-interface",
    "pipsi": "https://github.com/mitsuhiko/pipsi",
    "pipx": "https://github.com/pipxproject/pipx",
    "piwikapi": "https://github.com/piwik/piwik-python-api",
    "pkgwat-api": "https://github.com/fedora-infra/pkgwat.api",
    "PrettyTable": "https://code.google.com/p/prettytable",
    "prometheus-client": "https://github.com/prometheus/client_python",
    "prometheus_client": "https://github.com/prometheus/client_python",
    "protobuf3-to-dict": "https://github.com/kaporzhu/protobuf-to-dict",
    "ProxyTypes": "https://github.com/peak-legacy/proxytypes",
    "punch.py": "https://github.com/lgiordani/punch",
    "py3bitbucket": "https://github.com/c9s/py-bitbucket",
    "pyarrow": "https://github.com/apache/arrow",
    "pycontracts": "https://github.com/andreacensi/contracts",
    "pyenchant": "https://github.com/pyenchant/pyenchant",
    "pyethash": "https://github.com/ethereum/ethash",
    "pygresql": "https://github.com/pygresql/pygresql",
    "pyhacrf-datamade": "https://github.com/datamade/pyhacrf",
    "pynvim": "https://github.com/neovim/pynvim",
    "pyperformance": "https://github.com/python/pyperformance",
    "pyrex": "https://www.csse.canterbury.ac.nz/greg.ewing/python/Pyrex/",
    "pyspark": "https://github.com/apache/spark",
    "pytesting-utils": "https://github.com/pytesting/utils",
    "python2-hwloc": "https://gitlab.com/guystreeter/python-hwloc",
    "python2-libnuma": "https://gitlab.com/guystreeter/python-libnuma",
    "python2-pythondialog": "https://sourceforge.net/projects/pythondialog",
    "python3-hwloc": "https://gitlab.com/guystreeter/python-hwloc",
    "python3-libnuma": "https://gitlab.com/guystreeter/python-libnuma",
    "python-package-sync-tool": "https://github.com/alex-ber/PythonPackageSyncTool",
    "python-pseudorandom": "https://github.com/smathot/python-pseudorandom",
    "pyvows": "https://github.com/heynemann/pyvows",
    "pywatchman": "https://github.com/facebook/watchman",
    "recurly": "https://github.com/recurly/recurly-client-python",
    "reportportal-client": "https://github.com/reportportal/client-Python",
    "repoze.lru": "https://github.com/repoze/repoze.lru",
    "robotframework-databaselibrary": "https://github.com/franz-see/Robotframework-Database-Library",
    "robotframework-selenium2library": "https://github.com/robotframework/Selenium2Library",
    "robotframework-seleniumlibrary": "https://github.com/robotframework/SeleniumLibrary",
    "robotframework-sshlibrary": "https://github.com/robotframework/SSHLibrary",
    "rope": "https://github.com/python-rope/rope",
    "ruamel-ordereddict": "https://sourceforge.net/projects/ruamel-ordereddict",
    "ruamel-yaml-clib": "https://sourceforge.net/projects/ruamel-yaml-clib",
    "ruamel-yaml": "https://sourceforge.net/projects/ruamel-yaml",
    "rwt": "https://github.com/jaraco/rwt",
    "rx": "https://github.com/ReactiveX/RxPY",
    "ruffus": "https://github.com/cgat-developers/ruffus",
    "s3cmd": "https://github.com/s3tools/s3cmd",
    "sailthru-client": "https://github.com/sailthru/sailthru-python-client",
    "scaleapi": "https://github.com/scaleapi/scaleapi-python-client",
    "scikits-datasmooth": "https://github.com/jjstickel/scikit-datasmooth",
    "scikits-sparse": "https://github.com/scikit-sparse/scikit-sparse",
    "semantic-version": "https://github.com/rbarrois/python-semanticversion",
    "shopifyapi": "https://github.com/Shopify/shopify_python_api",
    "shrub-py": "https://github.com/evergreen-ci/shrub.py",
    "simplegeneric": "https://github.com/peak-legacy/simplegeneric",
    "smsapi-client": "https://github.com/smsapi/smsapi-python-client",
    "sseclient-py": "https://github.com/mpetazzoni/sseclient",
    "suds-jurko": "https://bitbucket.org/jurko/suds",
    "SymbolType": "http://svn.eby-sarna.com/SymbolType/",
    "SymbolType": "http://svn.eby-sarna.com/SymbolType/",
    "tableauserverclient": "https://github.com/tableau/server-client-python",
    "tap-py": "https://github.com/python-tap/tappy",
    "tensorflow-transform": "https://github.com/tensorflow/transform",
    "termstyle": "https://github.com/gfxmonk/termstyle",
    "tg.devtools": "https://github.com/TurboGears/tg2devtools",
    "threevis": "https://www.graphics.rwth-aachen.de:9000/threevis/threevis",
    "tifffile": "https://github.com/cgohlke/tifffile",
    "tlslite": "https://github.com/trevp/tlslite",
    "torchtext": "https://github.com/pytorch/text",
    "torchvision": "https://github.com/pytorch/vision",
    "treeherder-client": "https://github.com/mozilla/treeherder",
    "Trellis": "http://svn.eby-sarna.com/Trellis/",
    "twitter-ads": "https://github.com/twitterdev/twitter-python-ads-sdk",
    "ulid-py": "https://github.com/ahawker/ulid",
    "uritemplate-py": "https://github.com/python-hyper/uritemplate",
    "utmp": "https://github.com/hjacobs/utmp",
    "uwsgi": "https://github.com/unbit/uwsgi",
    "uwsgitop": "https://github.com/xrmx/uwsgitop",
    "webstack-django-jwt-auth": "https://github.com/webstack/django-jwt-auth",
    "wsgiref": "http://svn.eby-sarna.com/wsgiref/",
    "yubico-client": "https://github.com/Kami/python-yubico-client",
    "yubico": "https://github.com/Kami/python-yubico-client",
}

exact = exact_metadata.copy()
exact.update(exact_fetched)

name_mismatch_fetched = {
    "antlr-python-runtime": "https://github.com/antlr/antlr3",
    "antlr4-python2-runtime": "https://github.com/antlr/antlr4",
    "antlr4-python3-runtime": "https://github.com/antlr/antlr4",
    "atlassian-jwt-auth": "https://github.com/atlassian/asap-authentication-python",
    "atpublic": "https://gitlab.com/warsaw/public",
    "augurlib": "https://github.com/augur-ecosystem/augur",
    "avalara": "https://github.com/avadev/AvaTax-REST-V2-Python-SDK",
    "avocado-framework": "https://github.com/avocado-framework/avocado",
    "awkward": "https://github.com/scikit-hep/awkward-array",
    "aws-cdk-assets": "https://github.com/aws/aws-cdk",
    "aws-cdk-core": "https://github.com/aws/aws-cdk",
    "aws-cdk-cx-api": "https://github.com/aws/aws-cdk",
    "aws-cdk-region-info": "https://github.com/aws/aws-cdk",
    "awscli-cwlogs": "https://github.com/aws/aws-cli",
    "aws-kinesis-agg": "https://github.com/awslabs/kinesis-aggregation",
    "aws-sam-translator": "https://github.com/awslabs/serverless-application-model",
    "azure": "https://github.com/Azure/azure-sdk-for-python",
    "barnum": "https://github.com/chris1610/barnum-proj",
    "bert-serving-server": "https://github.com/hanxiao/bert-as-service",
    "blurb": "https://github.com/python/core-workflow",
    "boto3-stubs": "https://github.com/vemel/mypy_boto3_builder",
    "buildbot-slave": "https://github.com/buildbot/buildbot",
    "buildbot-worker": "https://github.com/buildbot/buildbot",
    "c7n": "https://github.com/cloud-custodian/cloud-custodian",
    "category-encoders": "https://github.com/scikit-learn-contrib/categorical-encoding",
    "cauldron-notebook": "https://github.com/sernst/cauldron",
    "celery-redbeat": "https://github.com/sibson/redbeat",
    "cfn-flip": "https://github.com/awslabs/aws-cfn-template-flip",
    "cfscrape": "https://github.com/Anorov/cloudflare-scrape",
    "chart-studio": "https://github.com/plotly/plotly.py",
    "cmreshandler": "https://github.com/cmanaha/python-elasticsearch-logger",
    "cogapp": "https://github.com/nedbat/cog",
    "cupy-cuda100": "https://github.com/cupy/cupy",
    "datanommer-consumer": "https://github.com/fedora-infra/datanommer",
    "datanommer-models": "https://github.com/fedora-infra/datanommer",
    "ddg3": "https://github.com/jpetrucciani/python-duckduckgo",
    "doc-warden": "https://github.com/Azure/azure-sdk-tools",
    "dogpile": "https://bitbucket.org/zzzeek/dogpile.core",
    "dwave-qbsolv": "https://github.com/dwavesystems/qbsolv",
    "ebcdic": "https://github.com/roskakori/CodecMapper",
    "exifread": "https://github.com/ianare/exif-py",
    "extract-msg": "https://github.com/mattgwwalker/msg-extractor",
    "eyes-selenium": "https://github.com/applitools/eyes.sdk.python",
    "f5-sdk": "https://github.com/F5Networks/f5-common-python",
    "feather-format": "https://github.com/wesm/feather",
    "flask-opentracing": "https://github.com/opentracing-contrib/python-flask",
    "freecell_solver": "https://github.com/shlomif/fc-solve",
    "gapic-google-cloud-logging-v2": "https://github.com/googleapis/googleapis",
    "gax-google-logging-v2": "https://github.com/googleapis/googleapis",
    "gax-google-pubsub-v1": "https://github.com/googleapis/googleapis",
    "gfapi": "https://github.com/gluster/libgfapi-python",
    "google-api-core": "https://github.com/googleapis/google-cloud-python",
    "googleappenginecloudstorageclient": "https://github.com/GoogleCloudPlatform/appengine-gcs-client",
    "gviz-api": "https://github.com/google/google-visualization-python",
    "h2": "https://github.com/python-hyper/hyper-h2",
    "hpsklearn": "https://github.com/hyperopt/hyperopt-sklearn",
    "ibm-cloud-sdk-core": "https://github.com/IBM/python-sdk-core",
    "ibm-watson": "https://github.com/watson-developer-cloud/python-sdk",
    "igor": "http://git.tremily.us/?p=igor.git",
    "jdatetime": "https://github.com/slashmili/python-jalali",
    "keras-mxnet": "https://github.com/awslabs/keras-apache-mxnet",
    "lftools": "https://gerrit.linuxfoundation.org/infra/q/project:releng%252Flftools",
    "lit": "https://github.com/llvm/llvm-project",
    "locustio": "https://github.com/locustio/locust",
    "lomond": "https://github.com/wildfoundry/dataplicity-lomond",
    "mailjet-rest": "https://github.com/mailjet/mailjet-apiv3-python",
    "marionette-driver": "https://wiki.mozilla.org/Auto-tools/Projects/Marionette",
    "matrix-client": "https://github.com/matrix-org/matrix-python-sdk",
    "maxminddb": "https://github.com/maxmind/MaxMind-DB-Reader-python",
    "mercurial": "https://www.mercurial-scm.org/repo/hg",
    "mibian": "https://github.com/yassinemaaroufi/MibianLib",
    "minimal-snowplow-tracker": "https://github.com/fishtown-analytics/snowplow-python-tracker",
    "mozcrash": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozdebug": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozdevice": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozfile": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozinfo": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozinstall": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozleak": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozlog": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "moznetwork": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozprocess": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozprofile": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozrunner": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "mozversion": "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "msal": "https://github.com/AzureAD/microsoft-authentication-library-for-python",
    "mwlib.docbook": "https://github.com/pediapress/mwlib",
    "mwlib.ext": "https://github.com/pediapress/mwlib",
    "mwlib.xhtml": "https://github.com/pediapress/mwlib",
    "mws": "https://github.com/python-amazon-mws/python-amazon-mws",
    "mypy-boto3": "https://github.com/vemel/mypy_boto3_builder",
    "nox-automation": "https://github.com/theacodes/nox",
    "opencensus-ext-logging": "https://github.com/census-instrumentation/opencensus-python",
    "opencensus-ext-requests": "https://github.com/census-instrumentation/opencensus-python",
    "opencensus-ext-stackdriver": "https://github.com/census-instrumentation/opencensus-python",
    "opencensus-ext-threading": "https://github.com/census-instrumentation/opencensus-python",
    "opencv-contrib-python-headless": "https://github.com/skvark/opencv-python",
    "opencv-contrib-python": "https://github.com/skvark/opencv-python",
    "opencv-python-headless": "https://github.com/skvark/opencv-python",
    "oss2": "https://github.com/aliyun/aliyun-oss-python-sdk",
    "pathlib-mate": "https://github.com/MacHu-GWU/pathlib_mate-project",
    "pathspec": "https://github.com/cpburnz/python-path-specification",
    "powerline-status": "https://github.com/powerline/powerline",
    "psycopg2-binary": "https://github.com/psycopg/psycopg2",
    "publicsuffixlist": "https://github.com/ko-zu/psl",
    "pycdio": "http://git.savannah.gnu.org/cgit/libcdio/pycdio.git",
    "pydevd-pycharm": "https://github.com/fabioz/PyDev.Debugger",
    "pyficache": "https://github.com/rocky/python-filecache",
    "pygtkhelpers": "https://bitbucket.org/aafshar/pygtkhelpers-main",
    "pygtk": "https://gitlab.gnome.org/GNOME/pygobject",
    "pyjarowinkler": "https://github.com/nap/jaro-winkler-distance",
    "pymongo": "https://github.com/mongodb/mongo-python-driver",
    "pyramid-arima": "https://github.com/tgsmith61591/pyramid",
    "pyside2": "http://code.qt.io/cgit/pyside/pyside-setup.git",  # close enough
    "pytest-profiling": "https://github.com/manahl/pytest-plugins",
    "pytest-server-fixtures": "https://github.com/manahl/pytest-plugins",
    "pytest-shutil": "https://github.com/manahl/pytest-plugins",
    "python-geoip-geolite2": "https://github.com/mitsuhiko/python-geoip",
    "python-vlc": "https://git.videolan.org/?p=vlc/bindings/python.git",
    "redfish": "https://github.com/DMTF/python-redfish-library",
    "regex": "https://bitbucket.org/mrabarnett/mrab-regex",
    "rfpimp": "https://github.com/parrt/random-forest-importances",
    "robotframework-pabot": "https://github.com/mkorpela/pabot",
    "rpi-gpio": "https://sourceforge.net/projects/raspberry-gpio-python",
    "scikit-surprise": "https://github.com/NicolasHug/Surprise",
    "skl2onnx": "https://github.com/onnx/sklearn-onnx",
    "sk-video": "https://github.com/scikit-video/scikit-video",
    "space-tracer": "https://github.com/donkirkby/live-py-plugin",
    "spark-parser": "https://github.com/rocky/python-spark",
    "strsimpy": "https://github.com/luozhouyang/python-string-similarity",
    "swapper": "https://github.com/wq/django-swappable-models",
    "tagpy": "http://git.tiker.net/?p=tagpy.git",
    "telepathy": "https://github.com/hackliff/intuition-plugins",
    "teradatasql": "https://github.com/Teradata/python-driver",
    "terminalone": "https://github.com/MediaMath/t1-python",
    "tf-estimator-nightly": "https://github.com/tensorflow/estimator",
    "tf-nightly-2-0-preview": "https://github.com/tensorflow/tensorflow",
    "tf-nightly-gpu": "https://github.com/tensorflow/tensorflow",
    "tf-nightly": "https://github.com/tensorflow/tensorflow",
    "thLib": "http://work.thaslwanter.at/thLib/html/",
    "turbomail": "https://github.com/marrow/marrow.mailer",
    "umap-learn": "https://github.com/lmcinnes/umap",
    "urbanairship": "https://github.com/urbanairship/python-library",
    "urlgrabber": "http://yum.baseurl.org/gitwebd27a.html?p=urlgrabber",
    "wandb": "https://github.com/wandb/client",
    "wxpython": "https://github.com/wxWidgets/Phoenix",
    "xkcdpass": "https://github.com/redacted/XKCD-password-generator",
    "yara": "https://github.com/mjdorma/yara-ctypes",
    "yoyo-migrations": "https://bitbucket.org/ollyc/yoyo",
    "zc.buildout": "https://github.com/buildout/buildout",
    "ZODB3": "https://github.com/zopefoundation/zodb",
}

name_mismatch_metadata = {
    "3to2": "https://bitbucket.org/amentajo/lib3to2",
    "absl-py": "https://github.com/abseil/abseil-py",
    "acme": "https://github.com/letsencrypt/letsencrypt",
    "adal": "https://github.com/AzureAD/azure-activedirectory-library-for-python",
    "adapt-parser": "https://github.com/MycroftAI/adapt",
    "adns": "https://github.com/trolldbois/python3-adns",
    "adyen": "https://github.com/Adyen/adyen-python-api-library",
    "aliyun-python-sdk-core": "https://github.com/aliyun/aliyun-openapi-python-sdk",
    "allure-behave": "https://github.com/allure-framework/allure-python",
    "allure-pytest": "https://github.com/allure-framework/allure-python",
    "allure-python-commons": "https://github.com/allure-framework/allure-python",
    "amazon-kclpy": "https://github.com/awslabs/amazon-kinesis-client-python",
    "ansicolors": "https://github.com/jonathaneunice/colors",
    "ansicolors": "https://github.com/jonathaneunice/colors",
    "antlr3-python-runtime": "https://github.com/antlr/antlr3",
    "apertium_lint": "https://sourceforge.net/projects/apertium",
    "asyncio-nats-client": "https://github.com/nats-io/nats.py",
    "asyncio-nats-streaming": "https://github.com/nats-io/stan.py",
    "authheaders": "https://github.com/ValiMail/authentication-headers",
    "authres": "https://launchpad.net/authentication-results-python",
    "aws-cdk-aws-iam": "https://github.com/aws/aws-cdk",
    "aws-cdk-aws-kms": "https://github.com/aws/aws-cdk",
    "aws-cdk-aws-s3": "https://github.com/aws/aws-cdk",
    "aws-client": "https://github.com/collectrium/col-aws-clients",
    "awsiotpythonsdk": "https://github.com/aws/aws-iot-device-sdk-python",
    "awswrangler": "https://github.com/awslabs/aws-data-wrangler",
    "azure-cosmosdb-nspkg": "https://github.com/Azure/azure-cosmosdb-python",
    "azure-cosmosdb-table": "https://github.com/Azure/azure-cosmosdb-python",
    "azure-devops": "https://github.com/Microsoft/vsts-python-api",
    "azure-kusto-data": "https://github.com/Azure/azure-kusto-python",
    "azure-kusto-ingest": "https://github.com/Azure/azure-kusto-python",
    "azure-storage-common": "https://github.com/Azure/azure-storage-python",
    "azure-storage-nspkg": "https://github.com/Azure/azure-storage-python",
    "azureml": "https://github.com/Azure/Azure-MachineLearning-ClientLibrary-Python",
    "azureml-model-management-sdk": "https://github.com/Azure/azure-ml-api-sdk-python",
    "basho-erlastic": "https://github.com/basho/python-erlastic",
    "bert-build": "https://git.cornhooves.org/build-tools/bert",
    "bert-serving-client": "https://github.com/hanxiao/bert-as-service",
    "bert-tensorflow": "https://github.com/google-research/bert",
    "BetterErrorMessages": "https://github.com/SylvainDe/DidYouMean-Python",
    "better-exceptions-fork": "https://github.com/delgan/better-exceptions",
    "bigcode-fetcher": "https://github.com/tuvistavie/bigcode-tools",
    "boa-api": "https://github.com/boalang/api-python",
    "boto3-type-annotations-with-docs": "https://github.com/alliefitter/boto3_type_annotations",
    "bugzilla": "https://github.com/gdestuynder/simple_bugzilla",  # https://github.com/bugzilla/bugzilla
    "cangjie": "https://github.com/bigdata-ustc/XKT",
    "cassandra-driver": "https://github.com/datastax/python-driver",
    "Catwalk": "https://github.com/pedersen/tgtools",
    "cdecimal": "https://www.bytereef.org/mpdecimal",
    "certbot-nginx": "https://github.com/letsencrypt/letsencrypt",
    "clickhouse-cityhash": "https://github.com/xzkostyan/python-cityhash",
    "cloudfoundry-client": "https://github.com/antechrestos/cf-python-client",
    "codeclimate-test-reporter": "https://github.com/codeclimate/python-test-reporter",
    "compatibility-lib": "https://github.com/GoogleCloudPlatform/cloud-opensource-python",
    "compressed-segmentation": "https://github.com/janelia-flyem/compressedseg",
    "copr-common": "https://pagure.io/copr/copr",
    "coreapi": "https://github.com/core-api/python-client",
    "coreapi": "https://github.com/core-api/python-client",
    "cornice-sphinx": "https://github.com/Cornices/cornice.ext.sphinx",
    "cql": "https://code.google.com/a/apache-extras.org/p/cassandra-dbapi2",
    "cqlsh": "http://git-wip-us.apache.org/repos/asf/cassandra.git",
    "crontab": "https://github.com/josiahcarlson/parse-crontab",
    "cryptography-vectors": "https://github.com/pyca/cryptography",
    "cs.urlutils": "https://bitbucket.org/cameron_simpson/css",
    "ctypeslib": "https://svn.python.org/projects/ctypes",
    "datadog-checks-base": "https://github.com/DataDog/integrations-core",
    "dbt-snowflake": "https://github.com/fishtown-analytics/dbt",
    "debian": "https://github.com/ourway/iran",
    "dedupe-hcluster": "https://github.com/datamade/hcluster",
    "dialogflow": "https://github.com/dialogflow/dialogflow-python-client-v2",
    "discover": "https://code.google.com/p/unittest-ext",
    "django-autoslug-iplweb": "https://github.com/neithere/django-autoslug",
    "django-common-helpers": "https://github.com/tivix/django-common",
    "django-cpserver": "https://hg.piranha.org.ua/cpserver",  # lost domain
    "django-daterange-filter": "https://github.com/tzulberti/django-datefilterspec",
    "django-events": "https://github.com/skyl/django-eventpickrs",
    "django-extra-fields": "https://github.com/Hipo/drf-extra-fields",
    "django-genericforeignkey": "http://tracpub.yaco.es/djangoapps/wiki/GenericForeignKey",
    "django-instant-coverage": "https://github.com/colons/instant-coverage",
    "django-kongoauth": "https://github.com/SchoolOrchestration/kongoauth",
    "django-notifications-hq": "https://github.com/django-notifications/django-notifications",
    "django-oot": "http://tracpub.yaco.es/djangoapps/wiki/OOT",
    "django-ranged-response": "https://github.com/wearespindle/django-ranged-fileresponse",
    "django-registration-redux": "https://github.com/macropin/django-registration",
    "djangorestframework-jwt-custom-user": "https://github.com/RegioHelden/django-rest-framework-jwt",
    "django-restricted-resource": "https://git.linaro.org/lava/django-restricted-resource.git",
    "django-roa": "http://code.welldev.org/django-roa/wiki/Home",
    "django-statsd-mozilla": "https://github.com/django-statsd/django-statsd",
    "django-stdfile": "https://tracpub.yaco.es/djangoapps/wiki/StdFile",
    "django-storages-redux": "https://github.com/jschneier/django-storages",
    "django-vz-wiki": "https://github.com/jobscry/vz-wiki",
    "djay": "https://github.com/aleontiev/dj",
    "dm-sonnet": "https://github.com/deepmind/sonnet",
    "dm-tree": "https://github.com/deepmind/tree",
    "dns-lexicon": "https://github.com/AnalogJ/lexicon",
    "docxtpl": "https://github.com/elapouya/python-docx-template",
    "dohq-artifactory": "https://github.com/devopshq/artifactory",
    "dopamine-rl": "https://github.com/google/dopamine",
    "dpcontracts": "https://github.com/deadpixi/contracts",
    "duo-web": "https://github.com/duosecurity/duo_python",
    "dynamodb-encryption-sdk": "https://github.com/aws/aws-dynamodb-encryption-python",
    "edgegrid-python": "https://github.com/akamai-open/AkamaiOPEN-edgegrid-python",
    "efilter": "https://github.com/google/dotty",
    "elastic-apm": "https://github.com/elastic/apm-agent-python",
    "elastic-apm": "https://github.com/elastic/apm-agent-python",
    "elasticsearch-async": "https://github.com/elastic/elasticsearch-py-async",
    "elasticsearch-curator": "https://github.com/elastic/curator",
    "elementtidy": "http://effbot.org/zone/element-tidylib.htm",
    "entropy-store": "https://github.com/fusionapp/entropy",
    "esrally": "https://github.com/elastic/rally",
    "etcd3gw": "https://github.com/dims/etcd3-gateway",
    "faiss-cpu": "https://github.com/kyamagu/faiss-wheels",
    "fake-factory": "https://github.com/joke2k/faker",
    "fbprophet": "https://github.com/facebook/prophet",
    "fb-re2": "https://github.com/facebook/pyre2",
    "Flask-Security-Too": "https://github.com/jwag956/flask-security",
    "flatten-json": "https://github.com/amirziai/flatten",
    "flit-core": "https://github.com/takluyver/flit",
    "fluidity-sm": "https://github.com/nsi-iff/fluidity",
    "FormatBlock": "https://github.com/welbornprod/fmtblock",
    "freetype": "https://sourceforge.net/projects/indic-computing",
    "fs": "https://github.com/PyFilesystem/pyfilesystem2",
    "fsleyes-props": "https://git.fmrib.ox.ac.uk/fsl/fsleyes/props",
    "fsleyes-widgets": "https://git.fmrib.ox.ac.uk/fsl/fsleyes/widgets",
    "fs-s3fs": "https://github.com/PyFilesystem/s3fs",
    "fsspec": "https://github.com/intake/filesystem_spec",
    "ftputil": "https://ftputil.sschwarzer.net/",
    "galaxy-containers": "https://github.com/galaxyproject/galaxy",
    "galaxy-sequence-utils": "https://github.com/galaxyproject/sequence_utils",
    "galaxy-tool-util": "https://github.com/galaxyproject/galaxy",
    "galaxy-util": "https://github.com/galaxyproject/galaxy",
    "geoip": "https://github.com/maxmind/GeoIP-api-python",
    "geventhttpclient-wheels": "https://github.com/gwik/geventhttpclient",
    "gherkin-official": "https://github.com/cucumber/gherkin-python",
    "ghubunix": "https://github.com/BBloggsbott/ghub",
    "git-github-reflog": "https://github.com/criswell/github-reflog",
    "github-feedparser": "https://github.com/wadewilliams/github-atom-feed-parser",
    "gmpy2": "https://github.com/aleaxit/gmpy",
    "GnuPGInterface": "https://sourceforge.net/projects/py-gnupg",  # sf redirect
    "googleapis-common-protos": "https://github.com/googleapis/googleapis",
    "google-auth-httplib2": "https://github.com/GoogleCloudPlatform/google-auth-library-python-httplib2",
    "google-auth": "https://github.com/googleapis/google-auth-library-python",
    "google-auth-oauthlib": "https://github.com/GoogleCloudPlatform/google-auth-library-python-oauthlib",
    "googledatastore": "https://github.com/GoogleCloudPlatform/google-cloud-datastore",
    "google-endpoints-api-management": "https://github.com/cloudendpoints/endpoints-management-python",
    "google-endpoints": "https://github.com/cloudendpoints/endpoints-python",
    "google-gax": "https://github.com/googleapis/gax-python",
    "googlemaps": "https://github.com/googlemaps/google-maps-services-python",
    "google-python-cloud-debugger": "https://github.com/GoogleCloudPlatform/cloud-debug-python",
    "googletranslate": "https://github.com/weaming/py-googletrans",
    "gremlinpython": "https://github.com/apache/tinkerpop",
    "grpc-google-iam-v1": "https://github.com/googleapis/googleapis",
    "grpc-google-logging-v2": "https://github.com/googleapis/googleapis",
    "grpc-google-pubsub-v1": "https://github.com/googleapis/googleapis",
    "grpcio": "https://github.com/grpc/grpc",
    "guppy": "https://github.com/svenil/guppy-pe",
    "h2o": "https://github.com/h2oai/h2o-3",
    "h2o-pysparkling-2.1": "https://github.com/h2oai/sparkling-water",
    "h2o-pysparkling-2.2": "https://github.com/h2oai/sparkling-water",
    "h2o-pysparkling-2.3": "https://github.com/h2oai/sparkling-water",
    "h2o-pysparkling-2.4": "https://github.com/h2oai/sparkling-water",
    "hachoir-core": "https://bitbucket.org/haypo/hachoir",
    "hachoir-metadata": "https://bitbucket.org/haypo/hachoir",
    "hachoir-parser": "https://bitbucket.org/haypo/hachoir",
    "home-assistant-frontend": "https://github.com/home-assistant/home-assistant-polymer",
    "homeassistant": "https://github.com/home-assistant/core",
    "http-ece": "https://github.com/martinthomson/encrypted-content-encoding",
    "hubble-github": "https://github.com/jos-b/hubble",
    "hypothesis-pytest": "https://github.com/DRMacIver/hypothesis",  # old repo
    "ibm-db-sa": "https://github.com/ibmdb/python-ibmdb",
    "ibmiotf": "https://github.com/ibm-watson-iot/iot-python",
    "icinga2py": "https://github.com/tclh123/icinga2-api",
    "instana": "https://github.com/instana/python-sensor",
    "interpret-core": "https://github.com/interpretml/interpret",
    "intuit-oauth": "https://github.com/intuit/oauth-pythonclient",
    "ipmi": "https://github.com/zhao-ji/check_ipmi_sensor_v3",
    "isim": "https://github.com/dalemyers/xcrun",
    "j1m.sphinxautozconfig": "https://github.com/jimfulton/sphinxautozconfig",
    "javaobj-py3": "https://github.com/tcalmant/python-javaobj",
    "json-logic-qubit": "https://github.com/qubitproducts/json-logic-py",
    "jsonrpclib-pelix": "https://github.com/tcalmant/jsonrpclib",
    "jupyter-kernel-gateway": "https://github.com/jupyter-incubator/kernel_gateway",
    "jupyter-kernel-polymake": "https://github.com/sebasguts/jupyter-polymake",
    "jupyter-kernel-singular": "https://github.com/sebasguts/jupyter-singular",
    "kaggle": "https://github.com/Kaggle/kaggle-api",
    "keras2onnx": "https://github.com/onnx/keras-onnx",
    "kerberos": "https://github.com/apple/ccs-pykerberos",
    "kfp-server-api": "https://github.com/kubeflow/pipelines",
    "kivy-garden": "https://github.com/kivy-garden/garden",
    "kiwisolver": "https://github.com/nucleic/kiwi",
    "kubernetes": "https://github.com/kubernetes-client/python",
    "launchdarkly-server-sdk": "https://github.com/launchdarkly/python-server-sdk",
    "lark-parser": "https://github.com/erezsh/lark",
    "ldclient-py": "https://github.com/launchdarkly/python-server-sdk",
    "libarchive": "https://github.com/dsoprea/PyEasyArchive",
    "libclang-py3": "https://hg.anteru.net/python3-libclang",
    "libmagic": "https://bitbucket.org/xmonader/pymagic-dev",
    "librato-metrics": "https://github.com/librato/python-librato",
    "liccheck": "https://github.com/dhatim/python-license-check",
    "linaro-django-jsonfield": "https://launchpad.net/django-jsonfield",
    "linear-tsv": "https://github.com/solidsnack/tsv",
    "localstack-ext": "https://github.com/localstack/localstack",
    "log4tailer": "https://bitbucket.org/jordilin/alerta",
    "lxc-python2": "https://github.com/lxc/python2-lxc",
    "m3-cdecimal": "https://www.bytereef.org/mpdecimal",
    "magma-lang": "https://github.com/phanrahan/magma",
    "MapGitConfig": "https://github.com/zalando/python-gitconfig",
    "marbles.core": "https://github.com/twosigma/marbles",
    "marbles.mixins": "https://github.com/twosigma/marbles",
    "MDAnalysisTests": "https://github.com/mdanalysis/mdanalysis",
    "MDP": "https://sourceforge.net/projects/mdp-toolkit",
    "mdx_gh_links": "https://github.com/Python-Markdown/github-links",
    "medusa": "https://sourceforge.net/projects/oedipus",
    "mglob": "https://code.google.com/p/vvtools",
    "misspellings": "https://github.com/lyda/misspell-check",
    "mlperf-compliance": "https://github.com/mlperf/training",
    "monkeypatcher": "https://github.com/iki/monkeypatch",
    "msk": "https://github.com/MycroftAI/mycroft-skills-kit",
    "msm": "https://github.com/MycroftAI/mycroft-skills-manager",
    "msrest": "https://github.com/Azure/msrest-for-python",
    "murmurhash3": "https://github.com/veegee/mmh3",
    "mysql": "https://github.com/valhallasw/virtual-mysql-pypi-package",
    "mysql-python": "https://github.com/farcepest/MySQLdb1",
    "neovim": "https://github.com/neovim/python-client",
    "netstorageapi": "https://github.com/akamai/NetStorageKit-Python",
    "nflx-genie-client": "https://github.com/netflix/genie",
    "nimoy-framework": "https://github.com/browncoat-ninjas/nimoy",
    "nixio": "https://github.com/G-Node/nixpy",
    "node-semver": "https://github.com/podhmo/python-semver",
    "nose-parameterized": "https://github.com/wolever/parameterized",
    "numarray": "https://sourceforge.net/projects/numpy",
    "numpy-quaternion": "https://github.com/moble/quaternion",
    "oauth_provider": "http://code.welldev.org/django-oauth/",
    "oic": "https://github.com/OpenIDC/pyoidc",
    "onetoken": "https://github.com/1token-trade/onetoken-py-sdk",
    "opencensus-context": "https://github.com/census-instrumentation/opencensus-python",
    "opencensus-ext-azure": "https://github.com/census-instrumentation/opencensus-python",
    "opencensus-ext-grpc": "https://github.com/census-instrumentation/opencensus-python",
    "openexr": "http://www.excamera.com/sphinx/articles-openexr.html",
    "openshift": "https://github.com/openshift/openshift-restclient-python",
    "optimizely-sdk": "https://github.com/optimizely/python-sdk",
    "os_virtual_interfacesv2_python_novaclient_ext": "https://github.com/rackerlabs/os_virtual_interfacesv2_ext",
    "paperboy": "https://github.com/edgeflip/dispatch",
    "parameterizedtestcase": "https://github.com/msabramo/python_unittest_parameterized_test_case",
    "patch-ng": "https://github.com/conan-io/python-patch",
    "paypalrestsdk": "https://github.com/paypal/PayPal-Python-SDK",
    "pdbpp": "https://github.com/antocuni/pdb",
    "pdoc3": "https://github.com/pdoc3/pdoc",
    "perfume-bench": "https://github.com/leifwalsh/perfume",
    "phonenumberslite": "https://github.com/daviddrysdale/python-phonenumbers",
    "pip-ppm": "https://github.com/ekinertac/ppm",
    "pip-update": "https://github.com/gogobook/python-env-upgradable-packages-extractor",
    "pkgcheck-arch": "https://github.com/FFY00/pkgcheck",
    "plaidml-keras": "https://github.com/plaidml/plaidml",
    "poline": "https://github.com/riolet/pol",
    "precise-runner": "https://github.com/MycroftAI/mycroft-precise",
    "predicthq": "https://github.com/predicthq/sdk-py",
    "pretrainedmodels": "https://github.com/cadene/pretrained-models.pytorch",
    "prison": "https://github.com/betodealmeida/python-rison",
    "progressbar33": "https://github.com/germangh/python-progressbar",
    "proto-google-cloud-error-reporting-v1beta1": "https://github.com/googleapis/googleapis",
    "proto-google-cloud-logging-v2": "https://github.com/googleapis/googleapis",
    "proto-google-cloud-pubsub-v1": "https://github.com/googleapis/googleapis",
    "publicsuffix": "https://code.google.com/p/python-public-suffix-list",
    "pulsar-galaxy-lib": "https://github.com/galaxyproject/pulsar",
    "pulsectl": "https://github.com/mk-fg/python-pulse-control",
    "pure-protobuf": "https://github.com/eigenein/protobuf",
    "pur": "https://github.com/alanhamlett/pip-update-requirements",
    "pusher": "https://github.com/pusher/pusher-http-python",
    "pyaml": "https://github.com/mk-fg/pretty-yaml",
    "pybarcode": "https://bitbucket.org/whitie/python-barcode",
    "pybitbucket_fork": "https://github.com/guyzmo/pybitbucket",
    "pybitbucket": "https://bitbucket.org/atlassian/python-bitbucket",
    "pybloom": "https://github.com/jaybaird/python-bloomfilter",
    "pycocotools": "https://github.com/cocodataset/cocoapi",
    "pyculiar": "https://github.com/wdm0006/pyculiarity",
    "pydeck": "https://github.com/uber/deck.gl",
    "pydevd": "https://github.com/fabioz/PyDev.Debugger",
    "pydocumentdb": "https://github.com/Azure/azure-documentdb-python",
    "pydomo": "https://github.com/domoinc/domo-python-sdk",
    "pyfarmhash": "https://github.com/veelion/python-farmhash",
    "pyGRID": "https://github.com/OpenMined/PySyft",
    "pyinsane2": "https://github.com/openpaperwork/pyinsane",
    "pylama_gjslint": "https://github.com/klen/pylama",
    "pyload-ng": "https://github.com/pyload/pyload",
    "pymacaroons-pynacl": "https://github.com/matrix-org/pymacaroons",
    "pymatbridge": "https://github.com/arokem/python-matlab-bridge",
    "pynut2": "https://github.com/mezz64/python-nut2",
    "pyobjc-core": "https://github.com/ronaldoussoren/pyobjc",
    "pyopengl-accelerate": "https://sourceforge.net/projects/pyopengl",
    "pypakr-pkg": "https://github.com/ajanicij/pypakr",
    "pypd": "https://github.com/PagerDuty/pagerduty-api-python-client",
    "pypillowfight": "https://gitlab.gnome.org/World/OpenPaperwork/libpillowfight",
    "pyqt4": "https://www.riverbankcomputing.com/software/pyqt",
    "pyqt5": "https://www.riverbankcomputing.com/software/pyqt",
    "pyqt5-sip": "https://www.riverbankcomputing.com/software/sip",
    "pyros-genmsg": "https://github.com/ros/genmsg",
    "pyros-genpy": "https://github.com/ros/genpy",
    "PyRSS2Gen": "http://dalkescientific.com/Python/PyRSS2Gen.html",
    "pysnmp-mibs": "https://sourceforge.net/projects/pysnmp",
    "pytest-allure-adaptor": "https://github.com/allure-framework/allure-pytest",
    "pytest-fixture-config": "https://github.com/manahl/pytest-plugins",
    "pytest-git": "https://github.com/manahl/pytest-plugins",
    "pytest-listener": "https://github.com/manahl/pytest-plugins",
    "pytest-pyramid-server": "https://github.com/manahl/pytest-plugins",
    "pytest-reportportal": "https://github.com/reportportal/agent-python-pytest",
    "pytest-svn": "https://github.com/manahl/pytest-plugins",
    "pytest-verbose-parametrize": "https://github.com/manahl/pytest-plugins",
    "pytest-virtualenv": "https://github.com/manahl/pytest-plugins",
    "python-graph-core": "https://github.com/pmatiello/python-graph",
    "python-graph-dot": "https://github.com/pmatiello/python-graph",
    "python-owasp-zap-v2-4": "https://github.com/zaproxy/zap-api-python",
    "python-potr": "https://github.com/afflux/pure-python-otr",
    "pythonz-bd": "https://github.com/berdario/pythonz",
    "qdarkstyle": "https://github.com/ColinDuquesnoy/QDarkStyleSheet",
    "qiniu": "https://github.com/qiniu/python-sdk",
    "qt4reactor": "https://github.com/ghtdak/qtreactor",
    "rapid-framework": "https://github.com/BambooHR/rapid",
    "ravello-sdk": "https://github.com/ravello/python-sdk",
    "readability-lxml": "https://github.com/buriy/python-readability",
    "repoman-scm": "https://github.com/tuenti/python-repoman",
    "repostat": "https://github.com/JamesJeffryes/repo-stat-stash",
    "repoze.what.plugins.sql": "https://github.com/repoze/repoze.what-sql",
    "repoze.who.plugins.sa": "https://github.com/repoze/repoze.who-sqlalchemy",
    "requests-cache-latest": "https://github.com/reclosedev/requests-cache",
    "requests-opentracing": "https://github.com/opentracing-contrib/python-requests",
    "respite": "https://github.com/jgorset/django-respite",
    "rest_framework_auth0": "https://github.com/mcueto/djangorestframework-auth0",
    "robotremoteserver": "https://github.com/robotframework/PythonRemoteServer",
    "rules": "https://github.com/dfunckt/django-rules",
    "scspell3k": "https://github.com/myint/scspell",
    "sentry-sdk": "https://github.com/getsentry/sentry-python",
    "setuptools-lint": "https://github.com/johnnoone/setuptools-pylint",
    "sfctl": "https://github.com/Azure/service-fabric-cli",
    "sfmergeutility": "https://github.com/Azure/azure-cli-extensions",
    "sgmllib3k": "https://hg.hardcoded.net/sgmllib",
    "sidecar": "https://github.com/jupyter-widgets/jupyterlab-sidecar",
    "slackweb": "https://github.com/satoshi03/slack-python-webhook",
    "smdebug-rulesconfig": "https://github.com/awslabs/sagemaker-debugger-rulesconfig",
    "smdebug": "https://github.com/awslabs/sagemaker-debugger",
    "snowballstemmer": "https://github.com/snowballstem/snowball",
    "social-auth-app-django": "https://github.com/python-social-auth/social-app-django",
    "social-auth-app-flask": "https://github.com/python-social-auth/social-app-flask",
    "social-auth-app-flask-sqlalchemy": "https://github.com/python-social-auth/social-app-flask-sqlalchemy",
    "social-auth-core": "https://github.com/python-social-auth/social-core",
    "social-auth-storage-mongoengine": "https://github.com/python-social-auth/social-storage-mongoengine",
    "social-auth-storage-peewee": "https://github.com/python-social-auth/social-storage-peewee",
    "social-auth-storage-sqlalchemy": "https://github.com/python-social-auth/social-storage-sqlalchemy",
    "sparkdl": "https://github.com/allwefantasy/spark-deep-learning",
    "sq-blocks": "https://github.com/square/blocks",
    "sst": "https://launchpad.net/selenium-simple-test",
    "stormssh": "https://github.com/emre/storm",
    "stups-tokens": "https://github.com/zalando-stups/python-tokens",
    "stups-zign": "https://github.com/zalando-stups/zign",
    "subgrab": "https://github.com/RafayGhafoor/Subscene-Subtitle-Grabber",
    "suds-community": "https://github.com/suds-community/suds",
    "systemd-python": "https://github.com/systemd/python-systemd",
    "taskcluster-urls": "https://github.com/taskcluster/taskcluster-lib-urls",
    "tb-nightly": "https://github.com/tensorflow/tensorboard",
    "tdlib": "https://github.com/tdlib/td",
    "tester-coverage": "https://github.com/vb64/test.helper.coverage",
    "tg-ext-silverplate": "https://github.com/pedersen/tgtools",
    "tgmochikit": "https://sourceforge.net/projects/turbogears1",
    "tox-monorepo": "https://github.com/Azure/azure-sdk-tools",
    "transip": "https://github.com/benkonrath/transip-api",
    "trepan3k": "https://github.com/rocky/python3-trepan",
    "TurboCheetah": "https://sourceforge.net/projects/turbogears1",
    "TurboGears2": "https://github.com/Turbogears/tg2",
    "turbokid": "https://sourceforge.net/projects/turbogears1",
    "tvb-gdist": "https://github.com/the-virtual-brain/tvb-geodesic",
    "tweet-preprocessor": "https://github.com/s/preprocessor",
    "twitter.common.finagle-thrift": "https://github.com/twitter/commons",
    "twitter.common.lang": "https://github.com/twitter/commons",
    "twitter.common.rpc": "https://github.com/twitter/commons",
    "typing-extensions": "https://github.com/python/typing",
    "uamqp": "https://github.com/Azure/azure-uamqp-python",
    "uiautomation": "https://github.com/yinkaisheng/Python-UIAutomation-for-Windows",
    "ujson-bedframe": "https://github.com/esnme/ultrajson",
    "ujson": "https://github.com/ultrajson/ultrajson",
    "ujson-ia": "https://github.com/internetarchive/ultrajson",
    "ujson-x": "https://github.com/PawelTroka/ultrajson-x",
    "umalqurra": "https://github.com/tytkal/python-hijiri-ummalqura",
    "upyun": "https://github.com/upyun/python-sdk",
    "uszipcode": "https://github.com/MacHu-GWU/uszipcode-project",
    "utillib": "https://github.com/jor-/util",
    "vim-vint": "https://github.com/Kuniwak/vint",
    "vsts": "https://github.com/Microsoft/vsts-python-api",
    "weka-easypy": "https://github.com/weka-io/easypy",
    "wn": "https://github.com/nltk/wordnet",
    "woocommerce": "https://github.com/woocommerce/wc-api-python",
    "word2number": "https://github.com/akshaynagpal/w2n",
    "wordaxe": "https://sourceforge.net/projects/deco-cow",
    "ws4py": "https://github.com/Lawouach/WebSocket-for-Python",
    "XStatic-Angular-Schema-Form": "https://github.com/dmsimard/python-XStatic-Angular-Schema-Form-distgit",
    "XStatic-objectpath": "https://github.com/dmsimard/python-XStatic-objectpath-distgit",
    "XStatic-tv4": "https://github.com/dmsimard/python-XStatic-tv4-distgit",
    "yarn-api-client": "https://github.com/toidi/hadoop-yarn-api-python-client",
    "yolk3k": "https://github.com/myint/yolk",
    "youtube-dl-server": "https://github.com/jaimeMF/youtube-dl-api-server",
    "z3-solver": "https://github.com/Z3Prover/z3",
    "zabbix-api": "https://github.com/gescheit/scripts",
    "zc.recipe.egg": "https://github.com/buildout/buildout",
    "ZSI": "https://sourceforge.net/projects/pywebsvcs",
    "zvmcloudconnector": "https://github.com/mfcloud/python-zvm-sdk",
}

name_mismatches = mismatch = name_mismatch_metadata.copy()
mismatch.update(name_mismatch_fetched)

invalid = ["tf-nightly-2-0-preview"]

bad_metadata = [
    "2to3",
    "abclient",
    "affinity",
    "aiotask-context",
    "alooma",
    "Aman",
    "amazon-dax-client",
    "ana",
    "anaconda",
    "anitya",
    "anykeystore",
    "application",
    "aspects",
    "augeas",
    "augur",
    "awsebcli",
    "awshelpers",
    "azure-functions",
    "azureml-automl-core",
    "azureml-automl-runtime",
    "azureml-core",
    "azureml-dataprep-native",
    "azureml-dataprep",
    "azureml-defaults",
    "azureml-explain-model",
    "azureml-interpret",
    "azureml-mlflow",
    "azureml-pipeline-core",
    "azureml-pipeline-steps",
    "azureml-pipeline",
    "azureml-sdk",
    "azureml-telemetry",
    "azureml-train-automl-client",
    "azureml-train-automl-runtime",
    "azureml-train-automl",
    "azureml-train-core",
    "azureml-train",
    "azureml-train-restclients-hyperdrive",
    "azureml-widgets",
    "bencode",
    "bigsuds",
    "bindep",
    "BitBucket",
    "bitbucket-cli",
    "BitVector",
    "bobodoctestumentation",
    "bottle-sqlite",
    "bs4",
    "btrc",
    "bugjar",
    "bugsnag",
    "Camelot",
    "carton",
    "catkin-sphinx",
    "cd",
    "cerealizer",
    "Cheesecake",  # should be https://github.com/griggheo/cheesecake
    "chromedriver",
    "chubby",
    "chubby",
    "client",
    "cloudshell-automation-api",
    "cmsis-pack-manager",
    "cognite-logger",
    "comet-git-pure",
    "comet-ml",
    "config",
    "coverage_pth",
    "cronwatch",
    "cvxopt",
    "daemon",
    "darcsver",  # add https://tahoe-lafs.org/trac/darcsver
    "databricks",
    "databricks-connect",  # https://github.com/Menziess/databricks-connect
    "databricks-pypi",
    "databricks-pypi1",
    "databricks-pypi2",
    "databricks-pypi-extras",
    "dateutils",  # todo
    "dbf",
    "dbus-python",
    "dci-utils",
    "demjson",
    "dependency_management",  # recheck
    "dai-sgqlc-3-5",
    "dialog",
    "digitalocean",
    "distribute",
    "distutils",
    "divide-and-cover",
    "django-dajaxice",  # https://github.com/jorgebastida/django-dajaxice/issues/158
    "django-datagrid",  # same company produces https://github.com/agiliq/Django-parsley, sometimes found
    "django-hijack",
    "django-inline-ordering",  # uses 'https://code.google.com/p/django-grappelli'
    "django-lfstheme",  # should be 'https://github.com/diefenbach/lfs-theme'
    "django-myrecaptcha",  # https://bitbucket.org/Kizlum/django-myrecaptcha ; https://bitbucket.org/pelletier/django-myrecaptcha exists still
    "django-settings-toml",
    "django-staticmediamgr",
    "django-toolbelt",  # finds https://github.com/heroku/django-heroku
    "djtracker",
    "dmidecode",
    "dm.xmlsec-binding",  # no repo exists by https://github.com/d-maurer, https://github.com/ansible/dm.xmlsec.binding is best attempt
    "dns",
    "dogstatsd-python",
    "dopen",
    "DoubleMetaphone",
    "dox",
    "dox-mortar",
    "dtopt",  # https://svn.colorstudy.com/dtopt offline
    "Durus",
    "electrical-calendar",
    "encutils",  # Should be https://code.google.com/p/cssutils
    "enum",  # stdlib
    "ephemeral-port-reserve",
    "epl",
    "evtx",
    "fiat",
    "fig",
    "firkin",
    "flojay",
    "fpconst",
    # "galaxy" very slow https://github.com/jayvdb/pypidb/issues/85
    "gen3git",
    "getch",
    "github3py",
    "github-issues-tools",
    "github-release",
    "Gloo",
    "google",
    "google-cloud-dataflow",
    "graph-tool",  # https://github.com/count0/graph-tool
    "hurry-filesize",
    "hypothesis-pytest",  # now part of https://github.com/DRMacIver/hypothesis
    "IGitt",
    "imblearn",
    "inifile",
    "intel-openmp",  # lots of odd results
    "interfile",
    "interval",
    "ipympl",
    "ipysheet",
    "jieba3k",  # py3 branch
    "jupyterlab-pygments",
    "justbases",
    "kaa-base",
    "kaa-metadata",
    "keystoneclient",
    "kid",
    "kitchen",
    "kmod",
    "launchpad-nag",
    "lightstep",
    "locust",
    "logentries",
    "logger",
    "lpshipit",
    "ly",
    "mailgun",
    "mailman-web",  # git@gitlab.com:mailman/mailman-web in PyPI description not detected
    "mallard-ducktype",
    "marionette",
    "marshmallow-enum",
    "membrete",
    "microversion-parse",
    "mkl",  # intel proprietary
    "moksha-common",
    "moksha-hub",
    "mongoalchemy",
    "monthdelta",
    "Morfessor",  # needs scraping depth 2 to find https://github.com/aalto-speech/morfessor mentioned on http://morpho.aalto.fi/projects/morpho/morfessor2.html
    "mozterm",
    "msal-extensions",
    "multicoder",
    "multimethods",
    "myghty",  # replaced with mako; homepage redirects back to pypi
    "mysql-connector",
    "mysql-connector-python",
    "mysql-connector-python-rf",
    "neptune-client",  # Finds https://github.com/neptune-ml/neptune-contrib
    "netapp-lib",
    "newrelic",
    "nose-fixes",
    "nose-fixes",  # < 10 downloads per day
    "nose-xunitmp",
    "np-utils",
    "nvidia-ml-py",
    "nvidia-ml-py3",
    "onedrivesdk",
    "openoffice",
    "openopt",
    "OrangeAssassin",
    "ordereddict",  # TODO: give it an activestate url
    "os-diskconfig-python-novaclient-ext",
    "oslo-sphinx",
    "overtest",  # redis results skipped
    "ovirt-engine-sdk",
    "ovirt-engine-sdk-python",
    "p4python",
    "packit",
    "pam",
    "panda",
    "password",
    "performance",
    "petlink",
    "pigpio",
    "pipit",  # https://github.com/youngking/pipit
    "pinax",
    "pip-reqs",
    "Pivy",
    "PlanOut",
    "pomegranate",
    "pprint",
    "prctl",
    "PrintDebug",
    "PsyLab",
    "ptrace",
    "pubnub",  # todo
    "py1cmd",
    "pyalsa",  # http://cvs.bigasterisk.com/viewcvs/pyalsa/ offline
    "pybee",
    "pybibframe",
    "pychart",
    "pycmd",
    "pycpuinfo",
    "pyfim",
    "pyformance",
    "pygal",
    "pygments-ansi-color",
    "pyjwkest",
    "pykdtree",
    "pykerberos",
    "pylirc",
    "pylogo",  # https://svn.colorstudy.com/PyLogo offline
    "PyMOL",
    "pymol",  # https://github.com/schrodinger/pymol-open-source, not on PyPI
    "pypiwin32",
    "pyprimes",
    "PyPrint",
    "pyprocdev",
    "PyProtocols",  # http://svn.eby-sarna.com/PyProtocols/, now at https://github.com/PEAK-Legacy/PyProtocols
    "Pyro",
    "pyrpm",
    "pyside2",
    "pyside2",
    "pyspark-dist-explore",
    "pytest-coverage",
    "pytext",
    "python-axolotl-curve25519",
    "python-docs-theme",
    "python-graph",  # no files, https://github.com/Shoobx/python-graph/blob/master/core/setup.py
    "python-hglib",
    "python-igraph",  # finds https://github.com/pygobject/pycairo
    "python-mpd",
    "python-uinput",
    "pythonwhois",
    "pytictoc",
    "pytidylib",
    "pytorch",
    "pyudorandom",
    "pyVirtualize",
    "pyweblib",
    "pyzipcode",
    "qet_tb_generator",
    "quickproxy",
    "random2",
    "reddit",
    "replit",  # todo
    "reportlab",
    "requestsexceptions",
    "ReviewBoard",
    "rfc6555",  # 0.0.0 (fixme)
    "robinhood-aiokafka",
    "rook",
    "ruptures",
    "s3pip",
    "s3pypi",
    "saganvm",
    "salt",
    "sane",
    "schevodurus",  # https://github.com/11craft/schevogears
    "scm-git",
    "score.projects",
    "script",
    "sed",
    "selection",
    "setuptools_cython",
    "shiboken2",
    "shippo",
    "simplewrap",
    "singer-python",
    "skimage",
    "sklearn",
    "SmartlingApiSdk",
    "solartime",
    "spark",  # www.hugedomains.com (http://trac.pytan.com)
    "sphinxcontrib-infrae",
    "sq-native",
    "sqlalchemy-teradata",
    "stacks",
    "staff-graded-xblock",
    "stemming",  # todo
    "strict-rfc3339",
    "stscipython",
    "suds",
    "supercaptcha",
    "tableauhyperapi",
    "tambo",
    "Tempita",
    "tensorboard-plugin-wit",
    "tensorflow-addons",
    "termcolor",
    "testflo",
    "testinfiniteloop",
    "tgfastdata",
    "timedop",
    "toastcord",
    "ToscaWidgets",
    "tracer",
    "travis",
    "triforce",
    "troveclient",
    "try",
    "ts-flint",
    "tw.dynforms",
    "tw.forms",
    "typing",
    "typogrify-engineer",
    "unittest",
    "urlopen",
    "urlparse3",
    "usbinfo",
    "vigilance",
    "viper",  # 2011, governance
    "volatility",
    "waiting",
    "watson-machine-learning-client",  # todo
    "wave",
    "webpy",  # joe job
    "websocket",
    "webunit",
    "widgetsnbextension",
    "Willow",
    "wmi",
    "xmltramp",  # cert failure & no files
    "zbase32",
    "zmq",
    "zuul-sphinx",  # https://opendev.org/zuul/zuul-sphinx choosing https://opendev.org/zuul/zuul instead, or None
]

sometimes_wrong_result = [
    "pyfeyn",  # https://svn.hepforge.org/pyfeyn and hepforge.org is offline
]

wrong_result = []

missing_repos = [
    "aioeventlet",
    "alfredcmd",  # https://github.com/GustavoKatel/alfredcmd
    "azureml-model-management-sdk",  # https://github.com/Azure/azure-ml-api-sdk-python
    "baruwa",  # https://github.com/akissa/baruwa disappeared
    "bcdoc",
    "boletin",  # https://tracpub.yaco.es/djangoapps/wiki/Boletin
    "clint2",  # https://github.com/myusuf3/clint
    "coding",  # https://bitbucket.org/krp/coding, http://code.activestate.com/pypm/coding/  by K Richard Pixley
    "curl2requests",  # https://github.com/a4m/curl2requests
    "datarobot",  # https://github.com/datarobot/public_api_client
    "dedupe-variable-datetime",  # https://github.com/datamade/dedupe-variable-datetime
    "debian",
    "django-attention",  # https://bitbucket.org/zacharyvoase/django-attention disappeared
    "django-cpserver",  # https://hg.piranha.org.ua/cpserver archived at https://github.com/lincolnloop/django-cpserver ?
    "django-debug",  # https://bitbucket.org/lorien/django-debug disappeaed
    "django-genericforeignkey",  # https://tracpub.yaco.es/..
    "django-livevalidation",
    "django-oot",  # https://tracpub.yaco.es/..
    "django-pingback",  # https://hg.piranha.org.ua/django-pingback
    "django-quickedit",
    "django-reporter",
    "django-retracer",
    "django-simple-cart",  # https://bitbucket.org/lorien/django-simple-cart
    "django-simple-news",
    "django-sproutcore",  # lost repo? https://github.com/sproutcoreisyournewbicycle/django-spoutcore has no setup.py check it even has a setup.py
    "django-stdfile",  # https://tracpub.yaco.es/djangoapps/wiki/StdFile
    "dnsproxy",  # https://github.com/blurrcat/dnsproxy
    "docker-buildtool",  # https://github.com/openai/docker-buildtool
    "elftools",  # https://github.com/jme2041/elftools
    "emencia_django_newsletter",
    "flask-mail",
    "fontconfig",  # https://bitbucket.org/gm/fontconfig
    "galley",  # https://github.com/rackerlabs/galley
    "github-console",  # https://github.com/luoyeah/github-console
    "github-languages",  # https://github.com/upcwangying/github-languages-python
    "github-repo",  # https://github.com/looking-for-a-job/github-repo.py (user disappeared)
    "glob3",
    "hachoir-core",  # haypo
    "hachoir-metadata",  # haypo
    "hachoir-parser",  # haypo
    "hgdistver",  # https://bitbucket.org/RonnyPfannschmidt/hgdistver decommissioned due to setuptools_scm
    "http-client",  # https://github.com/huynhminhson/http_client
    "ipmi",
    "iptcinfo",  # https://bitbucket.org/gthomas/iptcinfo , history retained at https://github.com/20minutes/iptcinfo
    "kubetop",  # https://github.com/leastauthority.com/kubetop
    "logging",
    "magic",  # is a DreamHost 404 page
    "metafone",  # https://github.com/al45tair/metaphone, but this is a dead fork
    "nose-testconfig",  # https://bitbucket.org/jnoller/nose-testconfig history at https://github.com/singingwolfboy/nose-testconfig
    "octohot",  # https://github.com/Hotmart-Org/octohot
    "pdfpages",
    "pip-extras",  # https://github.com/russel/pip_extras
    "pip-upgrade",  # https://github.com/alistair-broomhead/pip_upgrade
    "pybarcode",  # https://bitbucket.org/whitie/python-barcode/ history preserved at https://github.com/emfcamp/python-barcode
    "pygradle",  # https://github.com/mwalkowski/pygradle
    "pypi-uploader",  # https://github.com/ignacysokolowski/pypi-uploader
    "pyrepo",  # https://github.com/dghubble/pyrepo
    "python3-xlib",  # https://github.com/LiuLang/python3-xlib
    "python-emailahoy3",  # https://github.com/febrezo/python-emailahoy3
    "python-json2yaml",  # https://github.com/appstore-zencore/python-json2yaml
    "pythonloc",  # https://github.com/pipxproject/pythonloc
    "pythonpy",  # https://github.com/Russell91/pythonpy
    "rxjson",  # https://github.com/spiral-project/rxjson
    "safe-cast",  # https://github.com/jefftune/safe-cast
    "Schevo",
    "schevodurus",
    "SchevoGears",
    "SchevoSql",
    "searchform",  # https://tracpub.yaco.es/djangoapps/wiki/SearchForm
    "sgmllib3k",  # https://hg.hardcoded.net/sgmllib
    "spotinst-agent",  # https://github.com/spotinst/spotinst-sdk-python/issues/65
    "structurizr",  # https://github.com/sixty-north/structurizr
    "svg",  # https://github.com/007gzs/svg dummy package
    "telepathy",  # https://github.com/hackliff/intuition-plugins
    "tempdir",  # https://bitbucket.org/another_thomas/tempdir
    "transhette",  # https://tracpub.yaco.es/djangoapps/wiki/Transhette
    "unboxftpd",  # rewritten by same author to be 'django-ftpserver'
    "update",
    "virtualenvwrapper-django",  # https://bitbucket.org/dhellmann/virtualenvwrapper.django see also https://github.com/tokibito/djangohosting/commit/d07f28d57b753304479c18666795cbd4c3be3bd0 ; history preserved at https://github.com/dgladkov/virtualenvwrapper.django
    "xunitparser",  # https://git.p.engu.in/laurentb/xunitparser
]

setuppy_mismatches = [
    "acme",  # part of certbot
    "adduserpath",
    "aliyun-",
    "allure-",
    "ansible-tower-cli",
    "antlr-",
    "antlr3-",
    "antlr4-",
    "apache-beam",
    "airtable-python-wrapper",
    "asgi-redis",
    "aspy.yaml",
    "aws-cdk-",
    "awscli-cwlogs",
    "aws-kinesis-agg",
    "azure-cosmosdb-nspkg",
    "bert-serving-client",
    "bert-serving-server",
    "bert-tensorflow",
    "better-exceptions-fork",
    "bigcode-",
    "blurb",
    "boto3-type-annotations",
    "boto3-type-annotations-with-docs",
    "brewer2mpl",  # github rename
    "buildbot-",
    "cangjie",  # fork of https://pypi.org/project/XKT/
    "catkin-pkg",
    "Catwalk",
    "certbot-",
    "certbot$",
    "chart-studio",
    "cheetah",
    "cl",
    "compatibility-lib",
    "cronwatch",
    "cryptography-vectors",
    "cupy-cuda100",
    "dataclasses-serialization",
    "datadog",
    "dbt-",
    "devpi-",
    "django-announcements",
    "django-autoslug",  # lots o forks
    "django-celery-email",
    "django-logicaldelete",  # pinax-models
    "django-notification",
    "django-pyodbc-azure",  # wrong branch
    "django-ranged-response",  # -> django-ranged-fileresponse
    "djangorestframework-jwt",
    "django-rest-polymorphic",
    "django-storages-redux",
    "doc-warden",  # azure
    "dotnetcore2",  # doesnt contain the Python wrapper
    "dnspython3",  # dnspython
    "drf-jwt",  # lots o forks
    "ebcdic",
    "elasticsearch2",
    "elasticsearch5",
    "elasticsearch6",  # same as elasticsearch
    "eyes-selenium",
    "facebookads",
    "fake-factory",  # renamed to faker
    "fastdiff",  # python/ subdir
    "fbprophet",
    "feather-format",
    "flake8-junit-report",
    "fluent",
    "freecell-solver",  # Python may not in https://github.com/shlomif/fc-solve
    "fysom",
    "galaxy-",
    "gapic-",
    "gax-",
    "gcloud-",
    "gcloud$",
    "GeoFinder",  # 'Setup.py' vs 'setup.py'
    "geventhttpclient-wheels",
    "girder-client",  # renamed to girder
    "gitdb2",  # 'mirror' of gitdb (i.e. they rolled it back into the old project, and maybe the repos should actually be different)
    "google-api-",
    "googleapis-",
    "googleappenginecloudstorageclient",
    "google-cloud-",
    "google-cloud$",
    "google-compute-",
    "googledatastore",
    "google-nucleus",
    "google-python-cloud-debugger",  # src/
    "googletranslate",  # wrong name in setup.py
    "graphql-core-next",  # old repo for graphql-core
    "gremlinpython",
    "grpc-google-",
    "grpcio-",
    "grpcio$",
    "h2o",
    "h2o-",
    "httpcore",  # renamed to httpx
    "http-ece",
    "husl",  # https://github.com/hsluv/hsluv-python/blob/master/setup.py uses hsluv, but hsluv is 404 on pypi
    "hypothesis",
    "ibm-db",
    "ibm-db-sa",
    "ibmiotf",
    "igraph",  # renamed to jgraph
    "interpret-community",
    "interpret-core",
    "jinja",
    "jinja2",  # https://github.com/pallets/jinja2-compat
    "jupyter-nbformat",
    "jupyterlab-launcher",  # redirects to jupyterlab_server
    "keras2onnx",
    "kfp-server-api",
    "kivy-garden",  # https://github.com/kivy-garden/garden/pull/34
    "ldclient-py",
    "letsencrypt",
    "lit",
    "localstack-ext",
    "mapgitconfig",  # this is now https://gitlab.com/jmcs/python-gitconfig/blob/master/setup.py , which does include Map
    "MDAnalysis",  # setup not in root
    "MDAnalysisTests",  # lives in MDAnalysis test suite
    "mlperf-compliance",
    "monkeypatcher",  # -> monkeypatch
    "moretools",
    "mwlib-",
    "mxnet",
    "mysqlclient",
    "neo4j-driver",
    "nflx-genie-client",
    "nimbusml",
    "nose-parameterized",
    "nox-automation",
    "numpy-stl",
    "open3d-python",  # Intel
    "opencensus-ext-",
    "opentelemetry-api",
    "pantsbuild-pants",
    "patch",
    "path-py",  # -> path
    "pattern3",
    "phonenumberslite",  # -> phonenumbers
    "pinax-notifications",
    "pip-requ",
    "plaidml",
    "ply",  # has no setup.py
    "proto-",
    "proto-google-",
    "pyarrow",
    "pybitbucket-fork",
    "pybuilder-pip-tools",
    "pycli",  # name lives in <src>/__init__.py
    "pycocotools",
    "pydeck",
    "pydevd-pycharm",
    "pydocumentdb",  # repo & project renamed to azure-cosmos
    "pydot2",  # folded back into pydot project
    "PyGrid",
    "pyinsane2",
    "pylama-gjslint",
    "pymorphy2-dicts",
    "pyobjc-",
    "pyopenssl",
    "pypic",
    "pyros-",  # stringops to create name
    "pyshark",  # setup in src
    "pyspark",
    "pysyft",
    "pytest-codestyle",  # renamed to pytest-pycodestyle
    "pytest-devpi-server",
    "pytest-fixture-config",
    "pytest-git",
    "pytest-listener",
    "pytest-profiling",
    "pytest-pyramid-server",
    "pytest-qt-app",
    "pytest-server-fixtures",
    "pytest-shutil",
    "pytest-svn",
    "pytest-verbose-parametrize",
    "pytest-virtualenv",
    "python-awk",
    "python-geoip-geolite2",
    "python-graph-core",
    "python-graph-dot",
    "python-owasp-zap-v2-4",
    "pythonz-bd",  # setup.py contains pythonz
    "pytorch-pretrained-bert",
    "pytorch-transformers",
    "pyuwsgi",
    "py-vapid",
    "pyviz",
    "pywatchman",
    "Qprompt",
    "rasa-nlu",  # merged into rasa
    "ray",
    "records",
    "referer-parser",
    "requests-cache-latest",  # copy of requests-cache
    "rfpimp",  # src dir
    "rosdistro",
    "rpm",
    "rwt",  # pip-run github rename
    "satyr",  # mentor
    "scikits-sparse",
    "scons",
    "sfctl",
    "sfmergeutility",
    "sfx-jaeger-client",
    "shadowsocks",  # https://github.com/shadowsocks/shadowsocks "removed according to regulations"
    "sk-video",
    "smmap2",  # 'mirror' of smmap
    "smsapi-client",
    "snowballstemmer",
    "sparkdl",
    "sseclient-py",
    "stompest",
    "stompest",
    "tb-nightly",
    "tdlib",
    "tensorflow-",
    "teradatasql",
    "tf-estimator-nightly",
    "tf-nightly",
    "tg-ext-silverplate",  # empty repo
    "torchgeometry",
    "tox-monorepo",
    "treeherder-client",
    "trollius",  # branch trollius
    "turbomail",
    "turicreate",  # src/python/
    "tvb-gdist",  # string is split in setup.py
    "twitter-common-",
    "typing-extensions",
    "ujson-",  # lots of forks
    "umalqurra",
    "uritemplate-py",
    "vsts",
    "vsts-",
    "webapp",
    "XStatic-Angular-Schema-Form",
    "XStatic-objectpath",
    "xstatic-term-js",
    "XStatic-tv4",
    "yowsup2",  # repo contains yowsup
    "z3-solver",  # src/api/python/setup.py
    "zabbix-api",
    "zetup",
    "zodb3",
]

setuppy_mismatches_xstatic = [
    "xstatic-angular-lrdragndrop",
    "xstatic-bootstrap-datepicker",
    "xstatic-datatables",
    "xstatic-hogan",
    "xstatic-jquery",
    "xstatic-jquery-ui",
    "xstatic-jquery-migrate",  # these use https://github.com/openstack/xstatic-angular-lrdragndrop/blob/master/xstatic/pkg/angular_lrdragndrop/__init__.py
    "xstatic-jquery-quicksearch",
    "xstatic-jquery-tablesorter",
    "xstatic-magic-search",
    "xstatic-rickshaw",
    "xstatic-smart-table",
    "xstatic-spin",
]

expected = {}
expected.update(exact)
expected.update(mismatch)

failures = bad_metadata + sometimes_wrong_result + wrong_result + missing_repos



import json
import re

import requests
from fake_useragent import UserAgent
from logging_helper import setup_logging
from requests.exceptions import TooManyRedirects
from socialregexes import identify
from socialregexes.socialregexes import definitions as social_definitions
from stdlib_list import stdlib_list

from ._cache import _check_url_domain, get_file_cache_session, get_timeout
from ._db import _fetch_mapping, add_mapping, db_clear, mappings
from ._exceptions import (
    IncompletePackageMetadata,
    InvalidPackage,
    InvalidPackageVersion,
    PackageWithoutFiles,
    PackageWithoutUrls,
    UnrecognisedStdlibBackport,
)
from ._github import normalize
from ._rtd import get_repo as get_rtd_repo
from ._rules import DefaultRule, _find_named_repo, rules
from ._scm_url_cleaner import SCMURLCleaner
from ._similarity import _compute_similarity, get_best_match
from ._stdlib import ALLOWED_STDLIB_BACKPORTS
from ._types import Email, Name, Text, Url, UrlSet, Webpage

try:
    import brotli
except ImportError:
    brotli = None


logger = setup_logging()

_stdlib_all = stdlib_list()

social_definitions.update(
    {
        "gitlab": re.compile(r"https?://(?:www\.)?gitlab.com/([a-zA-Z0-9-]+)$"),
        "identica": re.compile(r"https?://(?:www\.)?identi.ca/([a-zA-Z0-9-]+)$"),
        "delicious": re.compile(r"https?://(?:www\.)?delicious.com/([a-zA-Z0-9-]+)$"),
        "delicious-api-1": re.compile(
            r"https?://api.del.icio.us/v1/js/[a-z]*/([a-zA-Z0-9-]+)?"
        ),
        "delicious-api-2": re.compile(
            r"https?://api.del.icio.us/v2/json/([a-zA-Z0-9-]+)/"
        ),
        "flattr": re.compile(r"https?://(?:www\.)?flattr.com/@([a-zA-Z0-9-]+)$"),
        "google": re.compile(
            r"^https?://profiles\.google\.com/((?:\+[a-zA-Z0-9.]+)|[0-9]+)$"
        ),
        "gravatar": re.compile(r"https?://www.gravatar.com/avatar.php?gravatar_id="),
    }
)

_DEFAULT_HEADERS = {
    "User-Agent": UserAgent().google,
    "Accept": "text/html,text/plain,application/*;q=0.8,text/*;q=0.5",
    "Accept-Encoding": "br,gzip;q=0.9,deflate;q=0.8"
    if brotli
    else "gzip,deflate;q=0.9",
    "Accept-Language": "en-US,en;q=0.9",
}


class Converter(object):
    def __init__(
        self,
        session=None,
        repo_url="https://pypi.org/pypi",
        website_timeout=13,
        max_fetches=None,
        clear_db=False,
        store_fetch_list=False,
    ):
        self.session = session
        self.web_session = session
        self._cleaner = SCMURLCleaner()
        self.repo_url = repo_url
        self.website_timeout = website_timeout
        self._max_fetch_count = max_fetches
        self.max_distance = 0.2
        self._store_fetch_list = store_fetch_list
        self.headers = _DEFAULT_HEADERS
        if clear_db:
            db_clear()

    def _get(self, *path):
        if self.session is None:
            self.session = get_file_cache_session("json")
        try:
            return self.session.get(
                self.repo_url.rstrip("/") + "/" + "/".join(path),
                allow_redirects=True,
                headers=self.headers,
            )
        except TooManyRedirects:
            # Simply repeat https://github.com/ionrock/cachecontrol/issues/214
            return requests.get(
                self.repo_url.rstrip("/") + "/" + "/".join(path),
                allow_redirects=True,
                headers=self.headers,
            )

    def _get_package_json(self, name):
        response = self._get("{0}/json".format(name))
        if response.status_code == 404:
            raise InvalidPackage("Invalid package name {}".format(name))
        else:
            data = json.loads(response.text)
            if not data:
                raise InvalidPackage("Invalid package data for name {}".format(name))
            return data

    def _accept_url(self, rule, name, url):
        url = self._cleaner.get_root(url)
        if not url:
            return url

        rv = rule.reject_match(name, url)
        if isinstance(rv, str):
            return rv
        elif rv:
            return False

        return url

    def _group_urls(self, urls, name):
        matches = set()
        related = set()
        non_matches = set()
        for url in urls:
            if name in url:
                matches.add(url)
            elif "python" in url.lower():
                related.add(url)
            elif "download" in url.lower():
                related.add(url)
            else:
                non_matches.add(url)
        return list(matches) + list(related) + list(non_matches)

    def _exclude_non_urls(self, urls):
        return set(
            url
            for url in urls
            if not url.endswith(".gz")
            and not url.endswith(".tgz")
            and not url.endswith(".xz")
            and not url.endswith(".bz2")
            and not url.endswith(".zip")
            and not url.endswith(".jar")
            and not url.endswith(".rar")
            and not url.endswith(".exe")
            and url != "UNKNOWN"
            and "ntp.org" not in url  # ntplib code excerpt
        )

    def _normalise_urls(self, urls):
        new_urls = set()
        for url in urls:
            hash_pos = url.find("#")
            if hash_pos == -1:
                new_urls.add(url)
            elif url[hash_pos - 1:].startswith("/#!/"):
                # http://travis-ci.org/#!/ misspellings, yara
                new_urls.add(url.replace("/#!/", "/"))
            else:
                new_urls.add(url[:hash_pos])
        return new_urls

    def _check_metadata(self, data):
        project_info = data["info"]
        name = project_info["name"]
        email = project_info.get("author_email", project_info.get("maintainer_email"))
        if not email:
            raise IncompletePackageMetadata(
                "{} has no email in PyPI metadata".format(name)
            )
        if (
            email != "UNKNOWN"
            and "@" not in email
            and " at " not in email
            and "_at_" not in email
        ):
            raise InvalidPackage(
                '{} has invalid email "{}" in PyPI metadata'.format(name, email)
            )

        license = project_info.get("license")
        if not license:
            raise IncompletePackageMetadata(
                "{} has no license in PyPI metadata".format(name)
            )

        summary = project_info.get("summary")
        if not summary:
            raise IncompletePackageMetadata(
                "{} has no summary in PyPI metadata".format(name)
            )

    def _check_artifacts(self, data):
        name = data["info"]["name"]
        check = data.get("releases")
        if not check:
            raise PackageWithoutFiles(
                "{} has no releases in PyPI metadata".format(name)
            )

        version = data["info"].get("version")
        if not version:
            raise InvalidPackageVersion(
                "{} has no version in PyPI metadata".format(name)
            )
        if version == "0" or set(version) == {"0", "."}:
            raise InvalidPackageVersion(
                "{} has version {} in PyPI metadata".format(name, version)
            )

        check = data.get("urls")
        if not check:
            raise PackageWithoutFiles(
                "{} (version {}) has no files in PyPI metadata".format(name, version)
            )

    def _get_inputs(self, rule, name, data):
        project_info = data.get("info", {})
        if not project_info:
            raise PackageWithoutFiles("{} has no info in PyPI metadata".format(name))
        name = project_info.get("name")
        if not name:
            raise InvalidPackage('{} has no "name" in PyPI metadata'.format(name))

        inputs = []

        project_urls = urls = project_info.get("project_urls") or []
        if urls:
            urls = list(urls.values())

        for key, value in project_info.items():
            if key.endswith("_url") and value and value != "UNKNOWN":
                if value not in project_urls:
                    urls.append(value)

        if not project_urls:
            self._check_artifacts(data)

            normalised_name = normalize(name)
            if normalised_name in _stdlib_all and name not in ALLOWED_STDLIB_BACKPORTS:
                raise UnrecognisedStdlibBackport(
                    "Package name {} conflicts with stdlib and is not a known backport".format(
                        name
                    )
                )

        if urls:
            urls_pre_dns = self._exclude_non_urls(urls)
            urls = set(url for url in urls_pre_dns if _check_url_domain(url))
            if urls_pre_dns != urls:
                logger.info(
                    "urls with domains not resolving: {}".format(
                        set(urls_pre_dns) - urls
                    )
                )

        emails = self._split_emails(project_info)
        summary = project_info.get("summary", "")
        description = project_info.get("description", "")

        redirects = rule.url_redirects()
        if redirects:
            for from_, to_ in redirects:
                if from_ and from_ in urls:
                    urls.remove(from_)
                    if to_:
                        urls.add(to_)
                elif to_:
                    urls.add(to_)

        if urls:
            inputs.append(UrlSet(urls))

        if summary:
            inputs.append(Text(summary, "summary"))

        if description:
            inputs.append(Text(description, "description"))

        for email in emails:
            inputs.append(Email(email))

        if not emails:
            inputs.append(Name(name))

        return inputs

    def _get_vcs_links(self, rule, name, data, inputs):
        normalized_name = normalize(name)
        results = []
        previous_results = None
        max_fetch_count = max(rule.fetch_count, self._max_fetch_count or 0)
        fetch_list = []
        seen_list = []
        fetch_count = 0
        queue = inputs[:]

        logger.debug("queue {}".format(queue))

        while queue:
            must_accept = False
            item = queue.pop(0)
            try:
                logger.debug(
                    "processing {}: {}".format(
                        item.__class__.__name__, item.source or item
                    )
                )
            except UnicodeEncodeError:
                logger.debug(
                    "processing {} (item decode problem)".format(
                        item.__class__.__name__
                    )
                )

            result_url = None
            if isinstance(item, UrlSet):
                urls = item.value
                urls = self._exclude_non_urls(urls)
                urls = self._normalise_urls(urls)

                if not urls:
                    continue

                seen_list += list(urls)

                fetch_urls = []
                result_urls = []

                for url in urls:
                    rv = self._accept_url(rule, name, url)
                    if rv:
                        results.append(rv)
                        result_urls.append(rv)
                    elif rv is not False:
                        fetch_urls.append(url)

                if len(result_urls) == 1 and len(fetch_list) <= 1:
                    results.append(result_urls[0])

                if result_urls:
                    logger.debug(
                        "--results from {}: {}".format(item.value, result_urls)
                    )
                else:
                    logger.debug("--none of this set: {}".format(item.value))

                logger.debug("queuing {}".format(sorted(fetch_urls)))
                for url in self._group_urls(fetch_urls, name):
                    queue.append(Url(url))

                logger.info(
                    "{}: from {} added urls {}".format(
                        name, item.source, sorted(result_urls)
                    )
                )

            elif isinstance(item, Text):
                if fetch_count > max_fetch_count:
                    logger.info("Not processing text from {}".format(item.source))
                    continue

                text = item.value
                if not text:
                    continue
                try:
                    urls = set(rule.link_extract(text, item.source))
                except Exception:
                    urls = []
                logger.debug(
                    "@@ ran {} on text size {} for {} urls !!".format(
                        rule.link_extract, len(text), len(urls)
                    )
                )
                logger.debug("extracted {}".format(sorted(urls)))
                queue.append(UrlSet(urls, item.source))

            elif isinstance(item, Email):
                email = item.value
                result_url = _find_named_repo(name, [email])
                must_accept = True

            elif isinstance(item, Name):
                item_name = item.value
                result_url = _find_named_repo(item_name)
                must_accept = True

            elif isinstance(item, Url):
                url = item.value

                if url in fetch_list:
                    logger.debug("queue loop skipping already fetched {}".format(url))
                    continue

                if url.startswith("https://github.com") or url.startswith(
                    "http://github.com"
                ):  # see github rule below
                    if ".github.com" not in url:
                        logger.debug("queue loop skipping github {}".format(url))
                        continue

                if fetch_count > max_fetch_count:
                    logger.debug(
                        "queue loop skipping >{}: {}".format(max_fetch_count, url)
                    )
                    continue

                user = identify(url)
                if user:
                    logger.info("{} detected as social for {}".format(url, user))
                    continue

                rv = rule.reject_url(name, url.lower())
                logger.debug("reject rule {}: {}".format(url, rv))
                if rv in ["", None]:
                    continue
                if rv is True:
                    continue

                if (
                    "github.com" in url and ".github.com" not in url
                ):  # yara, pykalman, membrete, zvmcloudconnector
                    logger.debug("queue loop skipping github v2 {}".format(url))
                    continue

                if self.web_session is None:
                    self.web_session = get_file_cache_session("web")

                if url.startswith("git://"):  # TODO: create tidy phase
                    url = url[6:]

                if not url.startswith("http://") and not url.startswith("https://"):
                    if "/" not in url:
                        url = "http://" + url + "/"
                    elif "://" not in url:
                        url = "http://" + url
                try:
                    logger.info("r {}".format(url))
                    r = self.web_session.get(
                        url, headers=self.headers, timeout=get_timeout(url)
                    )
                    logger.debug(
                        "r {}.url {} elapsed {}".format(
                            r.__class__.__name__, r.url, r.elapsed
                        )
                    )
                    logger.debug("r {} headers: {}".format(r.url, r.headers))
                    r.raise_for_status()
                except Exception as e:
                    logger.warning("{}: {}".format(url, e))
                    continue

                urls = []
                if r.url != url:
                    urls.append(r.url)

                if r.headers.get("X-RTD-Project"):
                    rtd_url = get_rtd_repo(r.headers.get("X-RTD-Project"))
                    if rtd_url:
                        logger.warning("{}: rtd {}".format(url, rtd_url))
                        urls.append(rtd_url)

                if urls:
                    queue.append(UrlSet(set(urls)))

                if not r.text:
                    logger.warning("{}: empty page".format(url))
                else:
                    queue.append(Webpage(r, url))
                    fetch_list.append(url)
                    fetch_count += 1

            if not result_url and results and results != previous_results:
                result_url = get_best_match(rule.match, results)
                previous_results = results[:]

            if result_url:
                score = _compute_similarity(rule.match, result_url)
                if must_accept or score < self.max_distance:
                    if self._store_fetch_list:
                        _fetch_mapping[normalized_name] = fetch_list

                    return result_url

            if not queue and not results:
                ph_url = "https://pythonhosted.org/{}/".format(name)
                if ph_url not in seen_list:
                    queue.append(Url(ph_url))

        logger.debug("fetched {}".format(fetch_list))
        if self._store_fetch_list:
            _fetch_mapping[normalized_name] = fetch_list

        if results:
            return get_best_match(rule.match, results)

    def _split_emails(self, project_info):
        emails = set(
            [project_info.get("author_email"), project_info.get("maintainer_email")]
        )
        emails = [
            email
            for email in emails
            if email and not email.startswith("UNKNOWN") and "@" in email
        ]
        logger.debug("emails {}".format(emails))
        split_emails = []
        for emails in emails:
            parts = emails.split(",")
            split_emails += [i.strip() for i in parts if i.strip() and "@" in i]
        return set(split_emails)

    def _add_mappings(self, names):
        for name in names:
            logger.debug("---- preloading {} ----".format(name))
            url = self.get_vcs(name)
            assert url
            add_mapping(name, url)
            logger.debug("---- preloaded {} = {} ----".format(name, url))

    def get_vcs(self, name):
        normalised_name = normalize(name)
        logger.info("looking up {}".format(name))
        cached_result = mappings.get(normalised_name)
        if cached_result:
            if isinstance(cached_result, str):
                return cached_result
            raise cached_result

        data = self._get_package_json(name)
        project_info = data.get("info", {})
        if not project_info:
            logger.debug(data)
            raise PackageWithoutFiles("{} has no info in PyPI metadata".format(name))

        name = project_info.get("name")
        if not name:
            raise InvalidPackage('{} has no "name" in PyPI metadata'.format(name))

        rule = rules.get(normalised_name)
        if not rule:
            rule = DefaultRule(name)
        logger.debug("rule {} {}".format(rule.__class__.__name__, rule))
        if rule.preload:
            self._add_mappings(rule.preload)

        inputs = self._get_inputs(rule, name, data)

        try:
            url = self._get_vcs_links(rule, name, data, inputs)
        except PackageWithoutUrls as err:
            logger.error(
                "_get_vcs_links({}) raised PackageWithoutUrls {!r}".format(name, err)
            )
            raise

        logger.debug("_get_vcs_links({}) returned .. {}".format(name, url))
        if not url:
            self._raise_no_result_exception(rule, name, data)

        if (
            normalised_name in _stdlib_all
            and normalised_name not in ALLOWED_STDLIB_BACKPORTS
        ):
            raise UnrecognisedStdlibBackport(
                "Package name {} ({}) conflicts with stdlib and is not a known backport".format(
                    name, url
                )
            )

        logger.debug("{}: found {}".format(name, url))

        if not rule.ignore_bad_metadata:
            self._check_artifacts(data)

        check_url = self._cleaner.get_root(url)
        assert check_url == url, "get_root({}) => {}".format(url, check_url)

        add_mapping(name, url)
        return url

    def _raise_no_result_exception(self, rule, name, data):
        project_info = data.get("info", {})
        if not project_info:
            raise PackageWithoutFiles("{} has no info in PyPI metadata".format(name))
        name = project_info.get("name")
        if not name:
            raise InvalidPackage('{} has no "name" in PyPI metadata'.format(name))

        self._check_artifacts(data)
        self._check_metadata(data)

        project_urls = urls = project_info.get("project_urls")
        urls = set(urls.values()) if urls else set()
        emails = self._split_emails(project_info)
        summary = project_info.get("summary", "")
        description = project_info.get("description", "")

        if not description:
            logger.info("summary and description missing from {}".format(project_info))
            raise IncompletePackageMetadata(
                "{} has no description in PyPI metadata".format(name)
            )

        raise PackageWithoutUrls("{} has no usable urls in PyPI metadata".format(name))



from __future__ import print_function

import sys

import click

from ._db import Database


@click.command()
@click.argument('name')
def cli(name):
    db = Database()
    try:
        url = db.find_project_scm_url(name)
        print(url)
    except Exception as e:
        error_msg = str(e)
        print(error_msg, file=sys.stderr)



from logging_helper import setup_logging

from ._similarity import normalize

logger = setup_logging()

mappings = {
    "tornado-proxy": "https://github.com/senko/tornado-proxy",
    "pbr": "https://opendev.org/openstack/pbr",
    "flux": "https://github.com/vmalloc/flux",
    "readthedocs-org": "https://github.com/rtfd/readthedocs.org",
}

reverse_mappings = {
    "https://opendev.org/openstack/project-config": "project-config",  # doesnt exist
    "https://github.com/hypothesisworks/hypothesis-python": "hypothesis",  # old url
    "https://github.com/logentries/lecli": "logentries-lecli",
    "https://github.com/senko/tornado-proxy": "tornado-proxy",
    "https://bitbucket.org/ubernostrum/django-contact-form": "django-contact-form",
    "https://gitlab.com/nekokatt/hikari.core": "hikari",
    "https://github.com/rtfd/readthedocs.org": "readthedocs.org",
    "https://sourceforge.net/projects/pywin32": "pywin32",
    "https://github.com/snide/sphinx_rtd_theme": "sphinx-rtd-theme",
    "https://opendev.org/openstack-dev/pbr": "pbr",
    "https://github.com/vmalloc/flux": "flux",
    "https://github.com/getslash/flux": "flux",
    "https://github.com/esnme/ultrajson": "ujson",
}

multipackage_repos = {
    "https://github.com/antlr/antlr3",
    "https://github.com/googleapis/googleapis",
    "https://wiki.mozilla.org/Auto-tools/Projects/Mozbase",
    "https://github.com/h2oai/sparkling-water",
    "https://github.com/manahl/pytest-plugins",
    "https://github.com/census-instrumentation/opencensus-python",
    "https://github.com/pyca/cryptography",
    "https://github.com/aws/aws-cli",
    "https://sourceforge.net/projects/turbogears1",
    "https://code.google.com/p/tgtools",
    "https://github.com/pedersen/tgtools",
    "https://code.google.com/p/ibm-db",
    "https://github.com/ibmdb/python-ibmdb",
    "https://github.com/googleapis/google-cloud-python",
    "https://github.com/marrow/mailer",
    "https://gitlab.gnome.org/GNOME/pygobject",
    "https://github.com/buildout/buildout",
    "https://www.riverbankcomputing.com/software/sip",
    "https://www.riverbankcomputing.com/software/pyqt",
    "https://www.bytereef.org/mpdecimal",
    "https://github.com/gfxmonk/termstyle",
    "https://github.com/OpenMined/PySyft",
    "https://github.com/plaidml/plaidml",
    "https://github.com/Kami/python-yubico-client",
    "https://sourceforge.net/projects/pyopengl",
    "https://github.com/mcfletch/pyopengl",
    "https://github.com/henry0312/pytest-codestyle",  # rename
    "https://github.com/Azure/azure-cli",
    "https://github.com/Microsoft/vsts-python-api",
    "https://github.com/Azure/azure-sdk-for-python",
    "https://github.com/Azure/azure-cosmosdb-python",
    "https://github.com/Azure/azure-kusto-python",
    "https://github.com/Azure/azure-storage-python",
    "https://github.com/fabioz/PyDev.Debugger",
    "https://github.com/galaxyproject/galaxy",
    "https://github.com/kyamagu/faiss-wheels",
    "https://github.com/unbit/uwsgi",
    "https://github.com/pmatiello/python-graph",
    "https://github.com/hanxiao/bert-as-service",
}
multipackage_repos = set(i.lower() for i in multipackage_repos)

_fetch_mapping = {}


def add_mapping(name, url):
    assert url is not None
    assert url != "http://sourceforge.net/p/turbogears1/code/"
    logger.info("Adding mapping {} = {}".format(name, url))
    normalised_name = normalize(name)
    mappings[normalised_name] = url
    if url.lower() not in multipackage_repos:
        # lower needed for hacking pycqa = PyCQA
        reverse_mappings[url.lower()] = (normalised_name, name)


def add_failed_mapping(name, err):  # pragma: no cover
    assert isinstance(err, Exception)
    logger.info("Adding failed mapping {} = {}".format(name, err))
    normalised_name = normalize(name)
    mappings[normalised_name] = err


def db_clear():  # pragma: no cover
    mappings.clear()
    reverse_mappings.clear()


class Database(object):
    def __init__(self, **kwargs):
        from ._pypi import Converter

        self._converter = Converter(**kwargs)
        self.website_timeout = self._converter.website_timeout
        self.projects = {}

    def find_project_scm_url(self, name):
        normalized_name = normalize(name)
        url = self._converter.get_vcs(normalized_name)
        self.projects[normalized_name] = url
        return url

    get_vcs = find_project_scm_url



from unittest_expander import expand, foreach

from pypidb._lp import _launchpad
from tests.data import exact, mismatch, missing_repos
from tests.utils import _TestBase, normalize, web_session

expected = {}
expected.update(exact)
expected.update(mismatch)


class TestLP(_TestBase):
    def test_pythonxy_archive(self):
        url = "https://code.launchpad.net/~pythonxy/+archive/pythonxy-devel"
        rv = _launchpad(url)
        self.assertEqual(rv, url.replace("code.", ""))

    def test_scm_bzr(self):
        url = self.converter.get_vcs("bzr")
        self.assertEqual(url, "https://launchpad.net/bzr")

    def test_BeautifulSoup(self):
        url = self.converter.get_vcs("BeautifulSoup")
        self.assertEqual(url, "https://launchpad.net/beautifulsoup")

    def test_beautifulsoup4(self):
        url = self.converter.get_vcs("beautifulsoup4")
        self.assertEqual(url, "https://launchpad.net/beautifulsoup")

    def test_testscenarios(self):
        url = self.converter.get_vcs("testscenarios")
        self.assertInsensitiveEqual(url, "https://launchpad.net/testscenarios")

    def test_testrepository(self):
        url = self.converter.get_vcs("testrepository")
        self.assertInsensitiveEqual(url, "https://launchpad.net/testrepository")
        # Also https://github.com/testing-cabal/testrepository

    def test_PyDispatcher(self):
        url = self.converter.get_vcs("PyDispatcher")
        self.assertInsensitiveEqual(
            url, "https://sourceforge.net/projects/pydispatcher"
        )
        # TODO: lower sf.net priority so answer is 'https://launchpad.net/pydispatcher

    def test_versiontools(self):
        url = self.converter.get_vcs("versiontools")
        self.assertInsensitiveEqual(url, "https://launchpad.net/versiontools")
        # TODO: should be enhanced to find https://github.com/zyga/versiontools which is
        # mentioned on https://code.launchpad.net/versiontools

    def test_jsonlib(self):
        url = self.converter.get_vcs("jsonlib-python3")
        self.assertInsensitiveEqual(url, "https://launchpad.net/jsonlib")

    def test_Parsley(self):
        url = self.converter.get_vcs("Parsley")
        self.assertInsensitiveEqual(url, "https://launchpad.net/parsley")

    def test_dkimpy(self):
        url = self.converter.get_vcs("dkimpy")
        self.assertInsensitiveEqual(url, "https://launchpad.net/dkimpy")

    def test_lazr_config(self):
        url = self.converter.get_vcs("lazr.config")
        self.assertInsensitiveEqual(url, "https://launchpad.net/lazr.config")

    def test_pkginfo(self):
        url = self.converter.get_vcs("pkginfo")
        self.assertInsensitiveEqual(url, "https://launchpad.net/pkginfo")

    def test_pyliblzma(self):
        url = self.converter.get_vcs("pyliblzma")
        self.assertInsensitiveEqual(url, "https://launchpad.net/pyliblzma")

    def test_wadllib(self):
        url = self.converter.get_vcs("wadllib")
        self.assertInsensitiveEqual(url, "https://launchpad.net/wadllib")

    def test_subunit(self):
        url = self.converter.get_vcs("python-subunit")
        self.assertInsensitiveEqual(url, "https://launchpad.net/subunit")

    def test_objgraph(self):
        url = self.converter.get_vcs("objgraph")
        self.assertEqual(url, "https://github.com/mgedmin/objgraph")

    def test_repoze_who(self):
        url = self.converter.get_vcs("repoze.who")
        self.assertInsensitiveEqual(url, "https://github.com/repoze/repoze.who")

    def test_zc_buildout(self):
        url = self.converter.get_vcs("zc.buildout")
        self.assertEqual(url, "https://github.com/buildout/buildout")

    def test_zc_recipe_egg(self):
        url = self.converter.get_vcs("zc.recipe.egg")
        self.assertEqual(url, "https://github.com/buildout/buildout")


@expand
class TestExplicit(_TestBase):

    expected = expected
    names = [name for name, url in expected.items() if "launchpad.net" in url]
    expected_failures = []

    @foreach(names)
    def test_package(self, name):
        expected = self.expected[name]
        try:
            url = self.converter.get_vcs(name)
        except Exception:
            if name in self.expected_failures:
                return
            raise

        if name in self.expected_failures:
            return

        self.assertIsNotNone(url)
        self.assertInsensitiveEqual(url, expected)

        normalised_name = normalize(name)

        if normalised_name in missing_repos:
            return
        r = web_session.get(url)
        r.raise_for_status()



import unittest

from unittest_expander import expand, foreach

from pypidb._exceptions import InvalidPackage

from tests.data import expected, failures
from tests.test_top import _all as _top_all
from tests.utils import _stdlib, _TestBase

try:
    from tests.datasets import get_fedora_packages, PortingdbLoader
except (SyntaxError, ImportError):
    raise unittest.SkipTest("Disabled on Python 2")

repeat_all = True
repeat_expected = True
repeat_failures = True


def _fetch_names():
    packages = get_fedora_packages()
    names = []
    for name in packages:

        if name in _stdlib:
            continue

        if not repeat_all and name.lower() in [i.lower() for i in _top_all]:
            continue

        if not repeat_expected and name.lower() in [i.lower() for i in expected]:
            continue

        if not repeat_failures and name.lower() in [i.lower() for i in failures]:
            continue

        names.append(name)

    return sorted(names)


_fedora_packages = _fetch_names()


@expand
class TestFedora(_TestBase):

    names = _fedora_packages
    expected_failures = [
        "abclient",
        "ana",
        "anykeystore",
        "augeas",
        "bigsuds",
        "bottle-sqlite",
        "catkin-sphinx",
        "Cerealizer",
        "certifi",
        "coverage_pth",
        "cvxopt",
        "dbf",
        "demjson",
        "dialog",
        "digitalocean",  # version 0
        "dmidecode",
        "dtopt",
        "fiat",
        "geoip",
        "graph-tool",
        "interfile",
        "justbases",
        "keystoneclient",  # version 0
        "kitchen",
        "kmod",
        "marshmallow-enum",
        "moksha.common",
        "moksha.hub",
        "mallard-ducktype",
        "nose-fixes",
        "openoffice",
        "openopt",
        "oslo.sphinx",  # no license
        "petlink",
        "pivy",
        "pyalsa",
        "pybugz",
        "pychess",
        "pycmd",
        "pycpuinfo",
        "pyfim",
        "pykdtree",  # no license
        "pymol",  # no files
        "pyprocdev",
        "pyro",
        "pyrpm",
        "py-spidev",
        "ptrace",
        "python-hglib",
        "python-mpd",
        "python-uinput",
        "pyvirtualize",
        "random2",
        "reportlab",
        "requestsexceptions",
        "script",  # no files
        "simplewrap",
        "suds",
        "tambo",
        "tempita",
        "termcolor",
        "tilestache",  # https://github.com/jayvdb/pypidb/issues/93
        "TroveClient",
        "zbase32",
        "zuul-sphinx",  # no license
    ]

    @foreach(names)
    def test_package(self, name):
        self._test_names([name])

    @foreach(PortingdbLoader._not_pypi)
    def test_invalid_name(self, name):
        if name not in ["docs"]:
            with self.assertRaises(InvalidPackage):
                self._get_scm(name)
        if not name.startswith("python-"):
            with self.assertRaises(InvalidPackage):
                self._get_scm("python-" + name)
        if not name.startswith("py") and name not in ["docs", "lxc"]:
            with self.assertRaises(InvalidPackage):
                self._get_scm("py" + name)
        if name.startswith("py") and name not in ["pybox2d", "pyev", "pysvn"]:
            with self.assertRaises(InvalidPackage):
                self._get_scm(name[2:])



from tests.utils import _TestBase


class TestShields(_TestBase):
    def test_invoke(self):
        url = self.converter.get_vcs("invoke")
        self.assertInsensitiveEqual(url, "https://github.com/pyinvoke/invoke")

    def test_pytest_testmon(self):
        url = self.converter.get_vcs("pytest_testmon")
        self.assertInsensitiveEqual(url, "https://github.com/tarpas/pytest-testmon")



from unittest_expander import expand, foreach

from tests.utils import _TestBase

# TODO: Use gitlab find https://opendev.org/explore/repos?q=requestsexceptions
openstack_none = ["microversion-parse", "requestsexceptions"]

openstack_opendev = [
    "jenkins-job-builder",
    "gear",
    "hacking",
    "mox3",
    "neutron-lib",
    "openstacksdk",
    "os-ken",
    "osc-lib",
    "pyeclib",
    "python-zaqarclient",
    "sphinx-feature-classification",
]

openstack_lp = ["lockfile"]

openstack_cgit = [
    "pbr",
    "bindep",
    "reno",
    "ldappool",
    "python-jenkins",
    "wsme",
    "glean",
    "bashate",
    "git-review",
    "python-rsdclient",
]

openstack_github = ["cliff"]


@expand
class TestOpenDev(_TestBase):

    expected_failures = ["bindep"]  # no license

    @foreach(openstack_opendev + openstack_cgit + openstack_github)
    def test_package(self, name):
        if name in self.expected_failures:
            return

        url = self.converter.get_vcs(name)
        self.assertIsNotNone(url)
        self.assertIn("//opendev.", url)
        self.assertIn("/" + name, url)


@expand
class TestOpenStackLaunchpad(_TestBase):
    @foreach(openstack_lp)
    def test_package(self, name):
        url = self.converter.get_vcs(name)
        self.assertIsNotNone(url)
        self.assertIn("//launchpad.", url)



import logging
from functools import partial

from cachetools import cached, LRUCache
from logging_helper import setup_logging

from ._cache import get_file_cache
from ._compat import PY2, logger_helper, parse_qs, urlsplit
from ._lp import _launchpad
from ._pagure import _pagure_io
from ._rtd import get_repo as _rtd_get_repo

logger = setup_logging()
if logger_helper and not PY2:
    autolog = logger_helper.LoggerHelper(logging.getLogger(__name__), logging.DEBUG)
else:  # pragma: no cover
    autolog = lambda x: x


def _match_hostname(url, condition, require_path=None, require_no_path=False):
    """require_path defaults to True unless match_subdomains is enabled."""
    scheme, _, other = url.partition(":")
    if scheme not in (
        "git",  # lxc-python2
        "git+https",  # asyncssh
        "http",
        "https",
        "svn",  # wsgiref
    ):
        return False

    if condition.startswith("http://"):
        condition = condition[7:]

    hostname, _, path = condition.partition("/")
    if ":" in hostname:
        hostname = hostname.split(":", 1)[0]

    if "." not in other:  # pragma: no cover
        return False  # '/dev/' in http://www.reportlab.com/

    other = other.lstrip("/")
    match_subdomains = hostname.startswith("*.")
    if match_subdomains:
        hostname = hostname[2:]

        subdomain, other = other.split(".", 1)
        if subdomain in ["www"]:
            logger.debug("url {} subdomain www".format(url))
            return False
    if not other.startswith(hostname):
        return None

    if require_path is None:
        require_path = not match_subdomains

    # Require at least a suffix
    other = other[len(hostname) :]
    other = other.lstrip("/")
    if not other:
        if require_no_path:
            return True

        if require_path:
            logger.debug("url {} no path".format(url))
            return False

    if path:
        if not other.startswith(path):
            logger.debug("url {} not path {}".format(url, path))
            return False

    return True


def _first_two_path(url):
    p = urlsplit(url)

    owner = p.path[1:]
    owner, _, other = owner.partition("/")
    repo, _, other = other.partition("/")

    if not owner or not repo:
        return

    return (owner, repo)


def _skip_x_path(url, skip=0, include_hostname=False, prefix=None):
    p = urlsplit(url)

    path = p.path[1:]
    while skip:
        _, _, path = path.partition("/")
        skip -= 1

    owner, _, path = path.partition("/")
    repo, _, other = path.partition("/")

    if not prefix:
        prefix = owner

    if include_hostname:
        return "https://{}/{}/{}".format(p.netloc, prefix, repo)
    else:
        return prefix, repo


def _skip_one_path(url, owner=None, hostname=None, scheme="https"):
    p = urlsplit(url)

    path = p.path[1:]
    if not path:
        return

    # Skip next 'foo/' in path
    _, _, path = path.partition("/")
    if not owner:
        owner, _, path = path.partition("/")

    repo, _, _ = path.partition("/")
    if hostname:
        return "{}://{}/{}/{}".format(scheme, hostname, owner, repo)

    return owner, repo


def _add_first_path(url, owner=None, hostname=None):
    p = urlsplit(url)

    path = p.path[1:]
    repo, _, other = path.partition("/")
    if hostname:
        return "https://{}/{}/{}".format(hostname, owner, repo)
    else:
        return (owner, repo)


def _hostname(url):
    p = urlsplit(url)
    return "https://{}/".format(p.netloc)


def _hostname_first_path(url):
    p = urlsplit(url)

    path = p.path[1:]
    repo, _, other = path.partition("/")
    return "https://{}/{}".format(p.netloc, repo)


def _hostname_two_paths(url, scheme="https", hostname=None, trailing=""):
    p = urlsplit(url)

    path = p.path[1:]
    path1, _, other = path.partition("/")
    path2, _, other = other.partition("/")
    if not path1 or not path2:
        return
    if not hostname:
        hostname = p.netloc
    return "{}://{}/{}/{}{}".format(scheme, hostname, path1, path2, trailing)


def _hostname_three_paths(url):
    p = urlsplit(url)

    path = p.path[1:]
    path1, _, other = path.partition("/")
    path2, _, other = other.partition("/")
    path3, _, other = other.partition("/")
    if not path1 or not path2 or not path3:  # pragma: no cover
        return
    return "https://{}/{}/{}/{}".format(p.netloc, path1, path2, path3)


def _subdomain_to_path(url, hostname=None, prefix=None, exclude=[], append_part=None):
    p = urlsplit(url)
    subdomain, _, domain = p.netloc.partition(".")
    if not hostname:
        hostname = domain
    if subdomain in exclude:
        return
    suffix = ""
    if append_part:
       if p.path[0] != "/":
           return
       parts = p.path[1:].split("/")
       suffix = "/" + parts[append_part - 1]
    if prefix:
        return "https://{}/{}{}{}".format(hostname, prefix, subdomain, suffix)
    return "https://{}/{}{}".format(hostname, subdomain, suffix)  # pragma: no cover


def _ghbtns(url):
    p = urlsplit(url)
    q = p.query
    if not p:  # pragma: no cover
        return
    params = parse_qs(q)
    if "user" not in params or "repo" not in params:
        return
    return params["user"][0], params["repo"][0]


def _subdomain_to_github(owner, url, exclude=[]):
    p = urlsplit(url)
    repo = p.netloc.split(".")[0]
    if repo in exclude:
        return False
    if (owner, repo) == ("apache", "qpid"):
        path = p.path
        if not path:  # pragma: no cover
            return
        path = path[1:]
        if not path:  # pragma: no cover
            return
        subproject, _, other = path.partition("/")
        if not subproject:  # pragma: no cover
            return
        repo = repo + "-" + subproject
    return (owner, repo)


def _github_io(url):
    p = urlsplit(url)
    owner = p.netloc.split(".")[0]
    logger.debug("_github_io {} {}".format(url, owner))
    if owner in [
        "api",
        "avatars",
        "buttons",  # plaidml-keras
        "developer",
        "enterprise",
        "help",
        "gist",
        "nodeload",
        "training",
        "user-images",
        "www",
    ]:
        return False
    path = p.path[1:]
    if not path:
        return
    repo, _, _ = path.partition("/")
    if not repo:  # pragma: no cover
        return
    # https://github.com/lipoja/URLExtract/issues/62#issuecomment-609275267
    if repo == "&":
        return
    if (owner, repo) in [("smartsheet-platform", "api-docs")]:
        return
    return (owner, repo)


def _strip_dot_git(url):
    pos = url.find(".git")
    if pos == -1:
        return url
    return url[:pos]


def _strip_after_dot_git(url):
    pos = url.find(".git")
    if pos == -1:
        return url
    return url[: pos + 4]


def _linaro_org(url):
    if "/git/" in url:
        return "https://git.linaro.org/{}/{}".format(*_skip_one_path(url))
    return url


def _all(url):
    return url


def _reject(url):
    return False


def _get_redirect_location(url):
    web_session = get_file_cache("web")
    if url.startswith("git+"):
        url = url[4:]
    r = web_session.get(url, allow_redirects=False)
    r.raise_for_status()
    assert "location" in r.headers
    assert r.headers["location"]
    url = r.headers["location"]
    if "opendev.org" in url and url.endswith("/"):
        url = url[:-1]
    return url


def _follow_rtd_get_repo(url):
    try:
        url = _get_redirect_location(url)
    except Exception:
        pass

    return _rtd_get_repo(url)


def _google_code(url):
    p = urlsplit(url)

    logger.debug("google code {}".format(url))

    if p.netloc == "issuetracker.google.com" and p.path.startswith("/code/"):
        p = urlsplit("https://code.google.com/" + p.path[6:])

    if p.netloc == "code.google.com":
        path = p.path[1:]
        if not path:  # pragma: no cover
            return
        part1, _, other = path.partition("/")

    else:
        url = _subdomain_to_path(url, prefix="p/", hostname="code.google.com")
        logger.debug("google code subdomain -> {}".format(url))
        p = urlsplit(url)

        path = p.path[1:]
        if not path:  # pragma: no cover
            return
        part1, _, other = path.partition("/")

    logger.debug("google code {} {}".format(part1, other))

    if part1 == "archive":
        part1, _, other = other.partition("/")

    if part1 == "a":
        return "https://code.google.com/a/{}".format(other)
    assert part1 == "p"
    project, _, other = other.partition("/")

    url = "https://code.google.com/p/{}".format(project)

    try:
        redirect = _get_redirect_location(url)
        if "code.google.com" not in redirect:
            return redirect
    except AssertionError:
        pass

    return url


def _telecommunity(url):
    if "setuptools" not in url:
        return _skip_one_path(hostname="svn.eby-sarna.com", scheme="http", url=url)


def _bitbucket(url):
    if url.startswith("https://drone.io/"):
        url = url = "https://" + url[len("https://drone.io/") :]

    url = _hostname_two_paths(url, hostname="bitbucket.org")
    if not url:
        return url

    try:
        web_session = get_file_cache("web")
        r = web_session.get(url, allow_redirects=False)
        content = r.content.decode("utf-8")
    except Exception as e:
        logger.info("bitbucket error {}".format(e))
        return url

    try:
        pos = content.find('It now lives at <a href="')
        if pos == -1:
            return url

        pos += len('It now lives at <a href="')
        end = content.find('"', pos)
        url = content[pos:end]
        logger.info("bitbucket redirecting to {}".format(url))
        url = url.replace(
            "https://sourceforge.net/p/", "https://sourceforge.net/projects/"
        )
    except Exception as e:
        logger.info("bitbucket error {}".format(e))
        raise

    if url.startswith("http://github.com"):
        url = url.replace("http://", "https://")
    elif url.startswith("https://sourceforge.net") and url[-1] == "/":
        url = url[:-1]
    elif url.startswith("https://hg.anteru.net") and url[-1] == "/":
        url = url[:-1]

    return url


def _zope_svn(url):
    p = urlsplit(url)
    if not p.path:  # pragma: no cover
        return False
    path = p.path.strip("/")
    if not path:  # pragma: no cover
        return False
    parts = path.split("/")
    if not parts:  # pragma: no cover
        return False
    name = parts[0]
    if not name:  # pragma: no cover
        return False
    if name == "zc.buildout":
        return ("buildout", "buildout")
    else:
        return ("zopefoundation", name)


def _gitlab(url, hostname=None):
    p = urlsplit(url)

    path = p.path[1:]
    if not path:  # pragma: no cover
        return
    if path.startswith("assets/"):
        return

    owner, _, path = path.partition("/")

    if not path:
        return
    for action in ["issues", "badges", "merge_requests"]:
        pos = path.find("/" + action)
        if pos != -1:
            path = path[:pos]
            break

    if not hostname:
        hostname = p.netloc
    return hostname, owner, path


_gitea = _hostname_two_paths  # gitea doesnt support groups


@autolog
def _github(url):
    try:
        rv = _first_two_path(url)
    except Exception as e:  # pragma: no cover
        logger.info("github split: {}".format(e))
        return False

    if not rv:
        return False
    owner, repo = rv

    if not owner or not repo:
        return False
    if owner in ["site", "users", "notifications", "customer-stories-feed"]:
        return False

    trimmed_repo, _, ext = repo.partition(".")
    if ext in ["git", "wiki"]:
        repo = trimmed_repo

    if (owner, repo) in [
        ("apache", "httpd"),  # pyside2
        ("reeset", "data_packs"),  # pybibframe
        ("typesafehub", "config"),  # strictyaml
        ("lightbend", "config"),  # strictyaml
        ("biolab", "orange"),  # https://github.com/biolab/orange3/pull/4344
        ("smartsheet-platform", "api-docs"),
    ]:
        return

    return owner, repo


_sf_rejects = [
    "apps",  # sf internal
    "djvu",  # python-djvulibre
    "lists",
    "ntlmaps",  # repoze.sphinx.autointerface
    "tidy",  # pytidylib
    "www",
]


def _sf(url):
    url = url.replace("sf.net", "sourceforge.net")

    if "turbogears1" in url:
        return url

    p = urlsplit(url)

    path = p.path[1:]

    if p.netloc == "lists.sourceforge.net":
        if path.startswith("lists/listinfo/"):
            list_name = path[len("lists/listinfo/") :]
        elif path.startswith("mailman/listinfo/"):
            list_name = path[len("mailman/listinfo/") :]
        else:
            assert False, path
        project_name = ""
        if list_name.endswith("-discuss"):
            project_name = list_name[: -len("-discuss")]
        if project_name:
            return "https://sourceforge.net/projects/{}".format(project_name)
        return
    elif p.netloc in ["sourceforge.net", "sf.net", "sourceforge.jp"]:
        if path in ["project/platformdownload.php", "project/showfiles.php"]:  # wordaxe
            url = _get_redirect_location(url.replace("http:", "https:"))
            p = urlsplit(url)
            path = p.path[1:]
    elif ".code." in p.netloc:
        pass
    else:
        subdomain = p.netloc
        subdomain = subdomain.replace(".sourceforge.net", "")
        subdomain = subdomain.replace(".sf.net", "")
        subdomain = subdomain.replace(".sourceforge.jp", "")
        logger.info("sf subdomain {} of {}".format(subdomain, url))
        if "." not in subdomain and subdomain not in _sf_rejects:
            return "https://sourceforge.net/projects/{}".format(subdomain)

    try:
        url = _skip_one_path(url, owner="projects", hostname="sourceforge.net")
    except Exception as e:  # pragma: no cover
        logger.warning("_sf exception: {}".format(e))
        return

    if not url:  # pragma: no cover
        logger.info("sf no url {}".format(url))
        return

    if not path:  # pragma: no cover
        logger.info("sf no path in {}".format(url))
        return

    name = url.rsplit("/", 1)[1]
    if not name:
        return False

    if name in _sf_rejects:
        logger.info("sf rejected {} in {}".format(name, _sf_rejects))
        return

    logger.info("accepted {} not in {}".format(name, _sf_rejects))

    return url


class SCMURLCleaner(object):
    fixers = {
        "raw.githubusercontent.com/": _first_two_path,
        "github.com": _github,
        "api.github.com/repos/": _skip_one_path,  # google-api-core
        "wiki.github.com": _first_two_path,  # pybrain
        "github.org": _github,  # tkreadonly, coverage-reporter
        "www.github.com": _github,
        "raw.github.com": _github,
        "travis-ci.org": _github,
        "travis-ci.com": _github,
        "www.travis-ci.org": _github,  # civisml-extensions
        "api.travis-ci.org": _github,
        "api.travis-ci.com": _github,
        "secure.travis-ci.org": _github,
        "bitbucket.org": _bitbucket,
        "www.bitbucket.org": _bitbucket,
        "gitlab.com": _gitlab,
        "www.gitlab.com": _gitlab,
        "rate.re/github/": _skip_one_path,  # docxtpl
        "codecov.io/gh/": _skip_one_path,
        "codecov.io/github/": _skip_one_path,
        "coveralls.io/github/": _skip_one_path,
        "coveralls.io/repos/github/": partial(_skip_x_path, skip=2),
        "coveralls.io/r/": _skip_one_path,
        "codeclimate.com/github/": _skip_one_path,
        "app.shippable.com": _skip_one_path,  # These should include /github
        "circleci.com/gh/": _skip_one_path,
        "nbviewer.jupyter.org/github/": _skip_one_path,
        "drone.io/bitbucket.org/": _bitbucket,
        "ci.appveyor.com/api/projects/status/github/": partial(_skip_x_path, skip=4),
        "requires.io/github/": _skip_one_path,
        "pyup.io/repos/github/": partial(_skip_x_path, skip=2),
        "img.shields.io/appveyor/ci/": partial(_skip_x_path, skip=2),
        "img.shields.io/codecov/c/github/": partial(_skip_x_path, skip=3),
        "img.shields.io/github/commit-activity/y/": partial(_skip_x_path, skip=3),
        "img.shields.io/github/license/": partial(_skip_x_path, skip=2),
        "img.shields.io/lgtm/alerts/g/": partial(_skip_x_path, skip=3),
        "img.shields.io/lgtm/grade/python/g/": partial(_skip_x_path, skip=4),
        "img.shields.io/travis/": _skip_one_path,
        "lgtm.com/projects/g/": partial(_skip_x_path, skip=2),
        "ghbtns.com/github-btn.html?": _ghbtns,
        "github.enthought.com/": partial(
            _add_first_path, owner="enthought"
        ),  # https://github.com/enthought/ets/issues/28
        "ssbjenkins.stsci.edu/job/STScI/job/": partial(
            _skip_x_path, skip=2, prefix="spacetelescope"
        ),
        "launchpad.net/": _launchpad,
        "www.launchpad.net/": _launchpad,
        "answers.launchpad.net/": _launchpad,
        "bazaar.launchpad.net/": _launchpad,
        "blueprints.launchpad.net/": _launchpad,
        "bugs.launchpad.net/": _launchpad,
        "code.launchpad.net/": _launchpad,
        "feeds.launchpad.net/": _launchpad,
        "translations.launchpad.net/": _launchpad,
        "pagure.io": _pagure_io,
        "savannah.gnu.org/projects/": _hostname_two_paths,
        "git.savannah.gnu.org/cgit/": _all,  # todo rewrite to savannah.gnu.org
        "git.sv.gnu.org": partial(
            _add_first_path, owner="projects", hostname="savannah.gnu.org"
        ),
        "savannah.nongnu.org/projects/": _hostname_two_paths,
        "sourceforge.net/project/platformdownload.php?group_id=": _sf,
        "sourceforge.net/project/showfiles.php?group_id=": _sf,
        "sourceforge.net/projects/": _sf,
        "sourceforge.jp/projects/": _sf,
        "sourceforge.net/p/": _sf,
        "apps.sourceforge.net/trac/": _sf,  # rewrite to sf.net/projects/
        "lists.sourceforge.net/lists/listinfo/": _sf,  # Cheetah
        "*.code.sourceforge.net/p/": _sf,
        "*.code.sf.net/p/": _sf,
        "sf.net/projects/": _sf,
        "*.sourceforge.net": _sf,
        "*.sf.net": _sf,
        # https://github.com/jayvdb/pypidb/issues/46
        "softwarefactory-project.io/r/": partial(
            _skip_x_path,
            skip=0, include_hostname=True, prefix="cgit",
        ),  # logreduce
        "*.softwarefactory-project.io": partial(
            _subdomain_to_path, prefix="cgit/"
        ),  # logreduce
        "softwarefactory-project.io/cgit/": _hostname_two_paths,
        "www.logilab.org/project/": _hostname_two_paths,
        "git-wip-us.apache.org/repos/": _all,
        "gitbox.apache.org/repos/": _all,
        "*.apache.org": partial(
            _subdomain_to_github,
            "apache",
            exclude=[
                "blogs",
                "community",
                "dist",
                "donate",
                "git-wip-us",
                "gitbox",
                "helpwanted",
                "httpd",
                "incubator",
                "issues",
                "jakarta",
                "labs",
                "people",
                "projects",
                "s",
                "status",
                "www",
            ],
        ),
        "pygobject.readthedocs.io": lambda url: "https://gitlab.gnome.org/GNOME/pygobject",
        "readthedocs.io/projects/": _rtd_get_repo,
        "readthedocs.org/projects/": _rtd_get_repo,
        "readthedocs.com/projects/": _rtd_get_repo,
        "*.readthedocs.io": _rtd_get_repo,
        "*.readthedocs.org": _rtd_get_repo,
        "*.readthedocs.com": _rtd_get_repo,
        "*.rtfd.org": _rtd_get_repo,
        "*.rtfd.io": _rtd_get_repo,
        "*.readthedocs-hosted.com": _follow_rtd_get_repo,
        "*.github.io": _github_io,
        "*.github.com": _github_io,
        "githubapi.com": _first_two_path,  # PyPI 'python-github' unaffiliated broken website
        "gitorious.org": _all,
        "gitorious.com": _all,
        "svn.turbogears.org": lambda url: "https://sourceforge.net/projects/turbogears1",
        "docs.turbogears.org/Turbo": lambda url: "https://sourceforge.net/projects/turbogears1",
        "docs.turbogears.org/tg": lambda url: "https://sourceforge.net/projects/turbogears1",
        "mercurial-scm.org/release/": lambda url: "https://www.mercurial-scm.org/repo/hg",
        "svn.timgolden.me.uk/wmi/trunk": lambda url: "https://github.com/tjguk/wmi",
        "dev.louiz.org/projects/slixmpp": lambda url: "https://lab.louiz.org/poezio/slixmpp",
        "docs.celeryproject.org/projects/": partial(
            _skip_one_path, owner="celery"
        ),  # This should translate second to https://github.com/celery/
        "docs.releng.linuxfoundation.org/projects/lftools": lambda url: "https://gerrit.linuxfoundation.org/infra/q/project:releng%252Flftools",
        "gerrit.linuxfoundation.org": _all,
        "code.google.com/a/": _google_code,
        "code.google.com/archive/a/": _google_code,
        "code.google.com/p/": _google_code,
        "code.google.com/archive/p/": _google_code,
        "issuetracker.google.com/p/": _google_code,
        "*.googlecode.com": _google_code,
        "git.openstack.org": _get_redirect_location,
        "gitlab.gnome.org": _gitlab,
        "foss.heptapod.net": _gitlab,
        "git.nzoss.org.nz": _gitlab,  # detect-delimiter
        "code.qt.io/cgit/": _strip_after_dot_git,
        "gitlab.freedesktop.org": _gitlab,
        "*.freedesktop.org/releases/": partial(
            _subdomain_to_path,
            hostname="gitlab.freedesktop.org",
            append_part=2,
        ),  # dbus-python
        "cgit.freedesktop.org/wiki/": _reject,
        "cgit.freedesktop.org": _hostname_two_paths,
        "git.tuxfamily.org/gitroot": _reject,
        "git.tuxfamily.org": _strip_after_dot_git,
        "salsa.debian.org": _all,
        "git.fmrib.ox.ac.uk": _gitlab,
        "gitlab.math.univ-paris-diderot.fr": _gitlab,
        "git.tremily.us/?p=": _strip_after_dot_git,  # also git://tremily.us/igor.git
        "hg.python.org/lookup": _reject,
        "hg.python.org/tracker": _hostname_two_paths,
        "hg.python.org": _hostname_first_path,
        "www.mercurial-scm.org/repo/hg": _all,
        "svn.python.org/projects/": _hostname_two_paths,
        "svn.zope.org": _zope_svn,
        "hg.tryton.org": _hostname_first_path,
        "*.tryton.org": partial(_subdomain_to_path, hostname="hg.tryton.org"),
        "hg.jcea.es": _hostname_first_path,
        "hg.hardcoded.net": _hostname_first_path,
        "hg.sschwarzer.net": _hostname_first_path,
        "*.sschwarzer.net": _hostname,
        "hg.mozilla.org": _all,
        "www.riverbankcomputing.com/hg/": _hostname_two_paths,
        "www.riverbankcomputing.com/software/": _hostname_two_paths,
        "hg.logilab.org": _gitlab,  # roughly correct
        "hg.piranha.org.ua": _hostname_first_path,
        "hg.anteru.net": _hostname_first_path,
        "www.tablix.org/~avian/git/": partial(
            _skip_x_path, skip=4, include_hostname=True
        ),  # Unidecode & publicsuffix are here, but they use GitHub now
        "yum.baseurl.org/gitwebd27a.html?p=": _strip_dot_git,
        "yum.baseurl.org/gitweb?p=": _strip_dot_git,
        "opendev.org": _gitea,
        "yourlabs.io": _gitlab,
        "gitlab.kitware.com": _gitlab,
        "git.cornhooves.org": _gitlab,
        "gitlab.lis-lab.fr": _gitlab,
        "forge-2.ircam.fr": _gitlab,  # pysndfile
        "www.graphics.rwth-aachen.de:9000": _hostname_two_paths,
        "www.saddi.com/software/": partial(_hostname_two_paths, trailing="/"),
        # svn.saddi.com': _hostname_first_path,
        "svn.hepforge.org": _hostname_first_path,  # 404 pyfeyn
        "http://peak.telecommunity.com/DevCenter/": _telecommunity,
        "http://svn.eby-sarna.com/svnroot/": _telecommunity,
        "http://svn.eby-sarna.com/": _all,
        "gitlab.mister-muffin.de": _hostname_two_paths,  # requires login?
        "dev.gajim.org": _hostname_two_paths,
        "lab.louiz.org/poezio/": _hostname_two_paths,
        "hg.sr.ht/": _hostname_two_paths,
        "reine.cmla.ens-cachan.fr/c.truong/ruptures": _gitlab,
        "git.tiker.net/": _all,
        "git.lxch.eu/": _github,  # remove once https://github.com/pcdummy/vcs-mirrors/pull/1 is released
        "git.framasoft.org": partial(_gitlab, hostname="framagit.org"),
        "framagit.org": _gitlab,
        "git.videolan.org/?p=": _strip_after_dot_git,
        "git.sysmocom.de": _all,
        "git.linaro.org/": _linaro_org,
        "el-tramo.be/git/": _all,
        "jelmer.uk/code/": _hostname_two_paths,
        "workbench.dachary.org/dachary/": _hostname_two_paths,
        "http://git.p.engu.in/": _hostname_two_paths,
        "code.dealmeida.net": partial(
            _add_first_path, owner="robertodealmeida", hostname="bitbucket.org"
        ),
        "http://git.gnu.org.ua/cgit/": _strip_after_dot_git,
        "tracpub.yaco.es/djangoapps/wiki/OOT": _all,  # offline?
        "trac.tools.ietf.org/tools/": _hostname_two_paths,  # xml2rfc
        "spacepants.org/src/": _hostname_two_paths,  # uses arch for vcs
        "www.owlfish.com/software/": _hostname_two_paths,  # wsgiutils
        "owlfish.com/software/": _hostname_two_paths,
        "http://zesty.ca/python/": _hostname_two_paths,
        "http://code.krypto.org/": _all,
        "guaix.fis.ucm.es/projects/": _hostname_two_paths,
        "people.csail.mit.edu/hubert/": _hostname_two_paths,
        "owlfish.com/software/simpleTAL/": _hostname_two_paths,
        "www.panix.com/~asl2/software/": _hostname_three_paths,
        "www.bytereef.org/mpdecimal": _hostname_first_path,
        "www.theblobshop.com/pymock": _all,  # https://github.com/jmyounker excludes pymock
        "effbot.org/zone/element-tidylib.htm": _all,
        "http://www.alcyone.com/software/": partial(_hostname_two_paths, scheme="http"),
        "work.thaslwanter.at/thLib/html/": _all,
        "http://dalkescientific.com/Python/PyRSS2Gen.html": _all,
        "https://developer.nvidia.com/nvidia-management-library-nvml": _all,
        "http://www.excamera.com/sphinx/articles-openexr.html": _all,
        "wiki.mozilla.org/Auto-tools/Projects/Mozbase": _all,
        "wiki.mozilla.org/Auto-tools/Projects/Marionette": _all,
        "www.cosc.canterbury.ac.nz/~greg/python/Pyrex": _get_redirect_location,
        "www.csse.canterbury.ac.nz/greg.ewing/python/Pyrex/": _all,
        "http://www.aaronsw.com/2002/xmltramp": _hostname_two_paths,
        "www.hpl.hp.com/personal/Yasushi_Saito/pychart": _all,
        "web.archive.org/web/20170225171913/https://gna.org/projects/pychart": _all,
        "code.welldev.org": _all,
        "web.archive.org/web/20130829003029/http://code.larlet.fr/django-oauth/wiki/Home": _all,
        "svn.wiretooth.com/svn/open/pypicache": _all,
        "http://www.jsnp.net/code/": _all,
        "web.archive.org/web/2009*/http://svn.wiretooth.com/svn/open/pypicache/": _all,
        "web.archive.org/web/20131016122638/http://vehq.ru/project/django-extended-attachments/": _all,
        "hg.vehq.ru": _all,  # TODO: test entries on http://hg.vehq.ru/index.html
        "tracpub.yaco.es/djangoapps/wiki/": _all,
        "git.liw.fi/cgi-bin/cgit/cgit.cgi/": _all,  # ttystatus https://liw.fi/ttystatus/
        "www.funaba.org/code": _all,
    }

    def _get_fixer(self, url):
        for host, func in self.fixers.items():
            if _match_hostname(url, host, require_path=False):
                return func

    @autolog
    def get_root(self, url):
        return self._get_root(url)

    @cached(cache=LRUCache(maxsize=32))
    def _get_root(self, url):
        func = self._get_fixer(url)
        if func:
            rv = func(url)
            if not rv:
                return rv
            if isinstance(rv, tuple):
                if len(rv) == 3:
                    hostname, owner, repo = rv
                else:
                    owner, repo = rv
                    hostname = "github.com"
                if not owner or not repo:
                    return

                at = repo.find("@")
                if at != -1:
                    repo = repo[:at]

                if (
                    repo.endswith(".git")
                    or repo.endswith(".svg")
                    or repo.endswith(".png")
                ):
                    repo = repo[:-4]

                # https://github.com/lipoja/URLExtract/issues/13
                repo = repo.strip(".")  # ebcdic / CodecMapper description
                # github-team-organizer -> https://github.com/theskumar/python-dotenv)
                repo = repo.strip(")")

                if "bitbucket.org/" in url:
                    return "https://bitbucket.org/{}/{}".format(owner, repo)
                elif "/gitlab.com/" in url:
                    return "https://gitlab.com/{}/{}".format(owner, repo)
                else:
                    return "https://{}/{}/{}".format(hostname, owner, repo)

            if url.startswith("http://github.com"):
                url = "https://github.com" + url[len("http://github.com") :]
            elif "opendev.org" in url and url.endswith("/"):
                url = url[:-1]
            return rv
